Unit BPInt;

{---------------------------------------------------------------------------}
{ Project: BP | Module BPInt - interface class for writing *.bpd files      }
{---------------------------------------------------------------------------}

 { Corrected 24.12.98 }
 { Rewritten 14-09-98 }
 { Rewritten 30-09-98 }
 { Modified  08-07-02 }
 { Modified  06-08-02 }

Interface

Uses Windows, Messages, SysUtils, Classes, Graphics,BPDef,BPStream,BPRename;

Type     TBPList = class(TComponent)
          private
            FTree:PBPTree;
            FHeader:TBPHeader;
            FLTTree:PBPLinkTypes;
            FNTTree:PBPNodeTypes;
            FOPTree:PBPOpTree;     { New: Operation Tree List }
            FEntryArray:PBPEntryArray;
            FCheckGraphArray:PBPCheckGraphArray;
            FNodes:PBPDNode;
            FLinks:PBPDLink;
{@@@}
            FIPNode:PBPInProcessNode;
{/@@@}
            FWNodes:PBPNode;
            FWLinks:PBPLink;
            FNodesCount:integer;
            FLinksCount:integer;
            FFileStream:TBPFileStream;
            FEditLevel:Boolean;
            FLevel:Integer;
            FParentLevel:Integer; { Parent Level }
            FDebug:Boolean;
            function    GetIndexedLink(index,lcount:integer;Links:PBPLink):PBPLink;
            procedure   RenameLinksNumbers;
            procedure   Copy;
            procedure   StoreEditingResults(wLevel:integer);
            procedure   RecreateNodesAndLinks;
            procedure   ClearNodes;
            procedure   ClearLinks;
            function    GetLinkTypeFrom(Flags:DWORD):TLinkType;
            function    GetLinkTypeTo(Flags:DWORD):TLinkType;
          public
            constructor Create(AOwner:TComponent);  override;
            destructor  Destroy; override;

            procedure   SetFileHeader(sModel,sCompany,sAuthor:ShortString;
                                      wDateTime:TDateTime;
                                      wRev:DWORD); { Set file header data }
            procedure   StartNewLevel(wLevel,wParent:Integer;sLevel,sCaption,sNotes:ShortString;
                                      wRev:DWORD;wStatus:TDraftStatus); { Start new level}
            function    SetCurrentLevel(wLevel:Integer):Boolean; { Query existing level,True if level exist }
            { New }
            procedure   ClearLinksTypes;
            procedure   ClearNodesTypes;
            procedure   ClearOperationTree; { New }
            procedure   ClearMSDAddon;
{@@@}
            procedure   ClearIPNode;
{/@@@}
            procedure   AddNodeType(index:DWORD;Caption:String);
            procedure   AddAdvNodeType(index:DWORD;Caption:String;FontColor,PaintColor,DrawColor,ShadowColor:TColor;ShadowStyle:DWORD);
            procedure   AddLinkType(index:DWORD;Caption:String);
            procedure   AddAdvLinkType(index:DWORD;Caption:String;Color:TColor;PenStyle:TPenStyle);
            procedure   AddOpNode(id,parent:DWORD;Caption:String;NodeIndex:DWORD;Detail:TStrings);
{@@@}
            procedure   AddIPNode(NodeCaption,Mech,Upr:ShortString;NodeNumber:DWORD;NodeIndex:DWORD);
            procedure   AddAdvIPNode(NodeCaption,Mech,Upr:ShortString;NodeNumber:DWORD;NodeIndex:DWORD;Comment,Path:ShortString);
{/@@@}
            { New }
            procedure   ClearAll; { Clear all tree structure }
            procedure   ClearCurrent; { Clear current nodes & links lists }
            procedure   AddNode(NodeCaption:ShortString;NodeNumber:DWORD;
                             NodeNumberCaption:TVeryShortString;ABCValue:Double;
                             NodeIndex:DWORD;Comment:ShortString); { Add node }
            procedure   IntAddLink(LinkFromCaption,LinkToCaption:ShortString;LinkGlobalCaption:TVeryShortString;
                                FromNode,ToNode:DWORD;
                                Flags:DWORD;
                                FromNumber,ToNumber:Integer;LinkIndex:DWORD;
                                FromPath,ToPath:ShortString); { Add Link }
            procedure   AddLink(LinkFromCaption,LinkToCaption:ShortString;LinkGlobalCaption:TVeryShortString;
                                FromNode,ToNode:DWORD;
                                FromLinkType,ToLinkType:TLinkType;
                                FromNumber,ToNumber:Integer;LinkIndex:DWORD;USINGFLAGS:DWORD;
                                FromPath,ToPath:ShortString); { Add link }
            procedure   SaveToFile(FileName:String); { Save result to file }
            procedure   LoadFromFile(Filename:String);
           published
          property NodesCount:integer read FNodesCount default 0;
          property LinksCount:integer read FLinksCount default 0;
          property Debug:Boolean read FDebug write FDebug;
         end;

procedure Register;


Implementation

 { Internal subroutines }

uses dialogs,debug;

{@@@}
 procedure   TBPList.ClearIPNode;
  begin
    if (FIPNode<>nil) then FreeMEm(FIPNode,sizeof(TBPInProcessNode)*FHeader.dwIPEntries);
    FIPNode:=nil;
    FHeader.dwIPEntries:=0;
  end;
{/@@@}

 procedure   TBPList.ClearMSDAddon;
  begin
    if (FEntryArray<>nil) then FreeMem(FEntryArray,sizeof(TBPEntryArray));
    if (FCheckGraphArray<>nil) then FreeMem(FCheckGraphArray,sizeof(TBPCheckGraphArray));
  end;

 procedure   TBPList.RenameLinksNumbers;
 Var tmpRename:TBPRename;
  begin
    tmpRename:=TBPRename.Create(FTree);
    tmpRename.RenameLinks;
    tmpRename.Destroy;
  end;

 function TBPList.GetIndexedLink(index,lcount:integer;Links:PBPLink):PBPLink;
  Var linkP:PBPLink;
  begin
    Result:=nil;
    if (index>=LCount) then Exit;
    linkP:=Links;
    if (index>0) then Inc(linkP,index);
    Result:=linkP;
  end;

 procedure   TBPList.StoreEditingResults(wLevel:integer);
 Var tmpTree:PBPTree;
     flag:Boolean;
  begin
    { Store editing results in tree }
    tmpTree:=FTree;
    Flag:=True;
    while (tmpTree<>nil) and (flag) do
     begin
       if (tmpTree^.dwLevel=wLevel) then
         begin
            Flag:=False;
            FEditLevel:=False;
            self.Copy;
            tmpTree^.dwNodesCount:=FNodesCount;
            tmpTree^.dwLinksCount:=FLinksCount;
            tmpTree^.dwNodesMemRequired:=FNodesCount*sizeof(TBPNode);
            tmpTree^.dwLinksMemRequired:=FLinksCount*sizeof(TBPLink);
            tmpTree^.lpNodes:=FWNodes;
            tmpTree^.lpLinks:=FWLinks;
            FWNodes:=nil;
            FWLinks:=nil;
            ClearCurrent;
         end;
       tmpTree:=tmpTree^.next;
     end;
  end;

 procedure   TBPList.ClearNodes;
 Var nodeP,nodeP1:PBPDNode;
  begin
      if (FNodes<>nil) then
      begin
        nodeP:=FNodes;
        while (nodeP<>nil) do
         begin
          nodeP1:=nodeP;
          nodeP:=nodeP^.next;
          FreeMem(nodeP1);
         end;
        FNodes:=nil;
      end;
  end;

 procedure   TBPList.ClearLinks;
 Var linkP,linkP1:PBPDLink;
  begin
    if (FLinks<>nil) then
      begin
        linkP:=FLinks;
        while (linkP<>nil) do
         begin
          linkP1:=linkP;
          linkP:=linkP^.next;
          FreeMem(linkP1);
         end;
        FLinks:=nil;
      end;
  end;

procedure   TBPList.ClearOperationTree;
Var tmpTree,tmpTree1:PBPOpTree;
 begin
   if (FOpTree<>nil) then
     begin
       tmpTree:=FOpTree;
       while (tmpTree<>nil) do
        begin
          if (tmpTree^.dwOpDetailCount>0) then
            FreeMem(tmpTree^.lpOpDetail,sizeof(TBPOpDetail)*tmpTree^.dwOpDetailCount);
          tmpTree1:=tmpTree;
          tmpTree:=tmpTree^.next;
          FreeMem(tmpTree1);
        end;
       FOpTree:=nil;
       FHeader.dwOpEntries:=0;
     end;
 end;

procedure   TBPList.ClearLinksTypes;
 begin
   if (FLTTree<>nil)
    then
     begin
       FreeMem(FLTTree,sizeof(TBPLinkTypes)*FHeader.dwLinksTypesCount);
       FLTTree:=nil;
       FHeader.dwLinksTypesCount:=0;
     end;
 end;

procedure   TBPList.ClearNodesTypes;
 begin
   if (FNTTree<>nil)
    then
     begin
       FreeMem(FNTTree,sizeof(TBPNodeTypes)*FHeader.dwNodesTypesCount);
       FNTTree:=nil;
       FHeader.dwNodesTypesCount:=0;
     end;
 end;

procedure   TBPList.AddOpNode(id,parent:DWORD;Caption:String;NodeIndex:DWORD;Detail:TStrings);
Var tmpTree,tmpTree1:PBPOpTree;
    pOpDetail,pOpDetail1:PBPOpDetail;
    detcount,i:Integer;
 begin
   if (Detail<>nil) then
          detcount:=Detail.Count
    else
          detcount:=0;
   if (FOpTree=nil) then
    begin
      tmpTree1:=PBPOpTree(AllocMem(sizeof(TBPOpTree)));
      tmpTree1^.dwDataOffset:=0;
      tmpTree1^.dwLevel:=id;
      tmpTree1^.dwParent:=parent;
      tmpTree1^.ssCaption:=caption;
      tmpTree1^.dwOpDetailCount:=detcount;
      tmpTree1^.dwNodeIndex:=NodeIndex;
      if (detcount>0) then
        begin
           pOpDetail:=PBPOpDetail(AllocMem(Sizeof(TBPOpDetail)*detcount));
           pOpDetail1:=pOpDetail;
           for i:=0 to (detcount-1) do
            begin
              pOpDetail1^.ssCaption:=Detail.Strings[i];
              inc(pOpDetail1);
            end;
           tmpTree1^.lpOpDetail:=pOpDetail;
        end
       else tmpTree1^.lpOpDetail:=nil;
      tmpTree1^.next:=nil;
      FOpTree:=tmpTree1;
      FHeader.dwOpEntries:=1;
    end
   else
    begin
      tmpTree1:=FOpTree;
      tmpTree:=nil;
      while (tmpTree1<>nil) do
        begin
          tmpTree:=tmpTree1;
          tmpTree1:=tmpTree1^.next;
        end;
      tmpTree1:=PBPOpTree(AllocMem(sizeof(TBPOpTree)));
      tmpTree1^.dwDataOffset:=0;
      tmpTree1^.dwLevel:=id;
      tmpTree1^.dwParent:=parent;
      tmpTree1^.ssCaption:=caption;
      tmpTree1^.dwOpDetailCount:=detcount;
      tmpTree1^.dwNodeIndex:=NodeIndex;
      if (detcount>0) then
        begin
           pOpDetail:=PBPOpDetail(AllocMem(Sizeof(TBPOpDetail)*detcount));
           pOpDetail1:=pOpDetail;
           for i:=0 to (detcount-1) do
            begin
              pOpDetail1^.ssCaption:=Detail.Strings[i];
              inc(pOpDetail1);
            end;
           tmpTree1^.lpOpDetail:=pOpDetail;
        end
       else tmpTree1^.lpOpDetail:=nil;
      tmpTree1^.next:=nil;
      tmpTree^.next:=tmpTree1;
      FHeader.dwOpEntries:=FHeader.dwOpEntries+1;
    end;
 end;

procedure   TBPList.AddAdvNodeType(index:DWORD;Caption:String;FontColor,PaintColor,DrawColor,ShadowColor:TColor;ShadowStyle:DWORD);
Var tmpPtr:PBPNodeTypes;
    tmpPtr1:PBPNodeTypes;
    i:Integer;
 begin
  {}
  if (FNTTree=nil) then
    begin
      FNTTree:=PBPNodeTypes(AllocMem(sizeof(TBPNodeTypes)));
      FNTTree^.dwNodeIndex:=index;
      FNTTree^.dwParent:=0;
      FNTTree^.dwID:=index;
      FNTTree^.ssCaption:=Caption;
      FNTTree^.dwFontColor:=FontColor;
      FNTTree^.dwPaintColor:=PaintColor;
      FNTTree^.dwDrawColor:=DrawColor;
      FNTTree^.dwShadowColor:=ShadowColor;
      FNTTree^.dwShadowStyle:=ShadowStyle;
      FNTTree^.dwRectStyle:=2;
      FHeader.dwNodesTypesCount:=1;
    end
  else
    begin
      tmpPtr:=PBPNodeTypes(AllocMem(sizeof(TBPNodeTypes)*(FHeader.dwNodesTypesCount+1)));
      Move(FNTTree^,tmpPtr^,sizeof(TBPNodeTypes)*FHeader.dwNodesTypesCount);
      tmpPtr1:=tmpPtr;
      for i:=1 to (FHeader.dwNodesTypesCount) do inc(tmpPtr1);
      tmpPtr1^.dwNodeIndex:=index;
      tmpPtr1^.dwID:=index;
      tmpPtr1^.dwParent:=0;
      tmpPtr1^.ssCaption:=Caption;
      tmpPtr1^.dwFontColor:=FontColor;
      tmpPtr1^.dwPaintColor:=PaintColor;
      tmpPtr1^.dwDrawColor:=DrawColor;
      tmpPtr1^.dwShadowColor:=ShadowColor;
      tmpPtr1^.dwShadowStyle:=ShadowStyle;
      tmpPtr1^.dwRectStyle:=2;
      FreeMem(FNTTree,sizeof(TBPNodeTypes)*FHeader.dwNodesTypesCount);
      FHeader.dwNodesTypesCount:=FHeader.dwNodesTypesCount+1;
      FNTTree:=tmpPtr;
      tmpPtr:=nil;
    end;
 end;

procedure   TBPList.AddNodeType(index:DWORD;Caption:String);
Var tmpPtr:PBPNodeTypes;
    tmpPtr1:PBPNodeTypes;
    i:Integer;
 begin
  {}
  if (FNTTree=nil) then
    begin
      FNTTree:=PBPNodeTypes(AllocMem(sizeof(TBPNodeTypes)));
      FNTTree^.dwNodeIndex:=index;
      FNTTree^.dwID:=index;
      FNTTree^.dwParent:=0;
      FNTTree^.ssCaption:=Caption;
      FNTTree^.dwFontColor:=clBlack;
      FNTTree^.dwPaintColor:=clWhite;
      FNTTree^.dwDrawColor:=clBlack;
      FNTTree^.dwShadowColor:=clBlack;
      FNTTree^.dwShadowStyle:=6;
      FNTTree^.dwRectStyle:=2;
      FHeader.dwNodesTypesCount:=1;
    end
  else
    begin
      tmpPtr:=PBPNodeTypes(AllocMem(sizeof(TBPNodeTypes)*(FHeader.dwNodesTypesCount+1)));
      Move(FNTTree^,tmpPtr^,sizeof(TBPNodeTypes)*FHeader.dwNodesTypesCount);
      tmpPtr1:=tmpPtr;
      for i:=1 to (FHeader.dwNodesTypesCount) do inc(tmpPtr1);
      tmpPtr1^.dwNodeIndex:=index;
      tmpPtr1^.dwID:=index;
      tmpPtr1^.dwParent:=0;
      tmpPtr1^.ssCaption:=Caption;
      tmpPtr1^.dwFontColor:=clBlack;
      tmpPtr1^.dwPaintColor:=clWhite;
      tmpPtr1^.dwDrawColor:=clBlack;
      tmpPtr1^.dwShadowColor:=clBlack;
      tmpPtr1^.dwShadowStyle:=6;
      tmpPtr1^.dwRectStyle:=2;
      FreeMem(FNTTree,sizeof(TBPNodeTypes)*FHeader.dwNodesTypesCount);
      FHeader.dwNodesTypesCount:=FHeader.dwNodesTypesCount+1;
      FNTTree:=tmpPtr;
      tmpPtr:=nil;
    end;
 end;

procedure   TBPList.AddAdvLinkType(index:DWORD;Caption:String;Color:TColor;PenStyle:TPenStyle);
Var tmpPtr:PBPLinkTypes;
    tmpPtr1:PBPLinkTypes;
    i:Integer;
 begin
  {}
    {}
  if (FLTTree=nil) then
    begin
      FLTTree:=PBPLinkTypes(AllocMem(sizeof(TBPLinkTypes)));
      FLTTree^.dwLinkIndex:=index;
      FLTTree^.dwID:=index;
      FLTTree^.dwParent:=0;
      FLTTree^.ssCaption:=Caption;
      FLTTree^.dwPenStyle:=PenStyle;
      FLTTree^.dwColor:=Color;
      FLTTree^.dbShowLinkType:=True;
      FHeader.dwLinksTypesCount:=1;
    end
  else
    begin
      tmpPtr:=PBPLinkTypes(AllocMem(sizeof(TBPLinkTypes)*(FHeader.dwLinksTypesCount+1)));
      Move(FLTTree^,tmpPtr^,sizeof(TBPLinkTypes)*FHeader.dwLinksTypesCount);
      tmpPtr1:=tmpPtr;
      for i:=1 to (FHeader.dwLinksTypesCount) do inc(tmpPtr1);
      tmpPtr1^.dwLinkIndex:=index;
      tmpPtr1^.dwID:=index;
      tmpPtr1^.dwParent:=0;
      tmpPtr1^.ssCaption:=Caption;
      tmpPtr1^.dwPenStyle:=PenStyle;
      tmpPtr1^.dwColor:=Color;
      tmpPtr1^.dbShowLinkType:=True;
      FreeMem(FLTTree,sizeof(TBPLinkTypes)*FHeader.dwLinksTypesCount);
      FHeader.dwLinksTypesCount:=FHeader.dwLinksTypesCount+1;
      FLTTree:=tmpPtr;
      tmpPtr:=nil;
    end;

 end;

procedure   TBPList.AddLinkType(index:DWORD;Caption:String);
Var tmpPtr:PBPLinkTypes;
    tmpPtr1:PBPLinkTypes;
    i:Integer;
 begin
  {}
    {}
  if (FLTTree=nil) then
    begin
      FLTTree:=PBPLinkTypes(AllocMem(sizeof(TBPLinkTypes)));
      FLTTree^.dwLinkIndex:=index;
      FLTTree^.dwID:=index;
      FLTTree^.dwParent:=0;
      FLTTree^.ssCaption:=Caption;
      FLTTree^.dwPenStyle:=psSolid;
      FLTTree^.dwColor:=clBlack;
      FLTTree^.dbShowLinkType:=True;
      FHeader.dwLinksTypesCount:=1;
    end
  else
    begin
      tmpPtr:=PBPLinkTypes(AllocMem(sizeof(TBPLinkTypes)*(FHeader.dwLinksTypesCount+1)));
      Move(FLTTree^,tmpPtr^,sizeof(TBPLinkTypes)*FHeader.dwLinksTypesCount);
      tmpPtr1:=tmpPtr;
      for i:=1 to (FHeader.dwLinksTypesCount) do inc(tmpPtr1);
      tmpPtr1^.dwLinkIndex:=index;
      tmpPtr1^.dwParent:=0;
      tmpPtr1^.dwID:=index;
      tmpPtr1^.ssCaption:=Caption;
      tmpPtr1^.dwPenStyle:=psSolid;
      tmpPtr1^.dwColor:=clBlack;
      tmpPtr1^.dbShowLinkType:=True;
      FreeMem(FLTTree,sizeof(TBPLinkTypes)*FHeader.dwLinksTypesCount);
      FHeader.dwLinksTypesCount:=FHeader.dwLinksTypesCount+1;
      FLTTree:=tmpPtr;
      tmpPtr:=nil;
    end;

 end;

 procedure   TBPList.RecreateNodesAndLinks;
 Var nodeP,oldNodeP:PBPDNode;
     nodeW:PBPNode;
     linkP,oldLinkP:PBPDLink;
     linkW:PBPLink;
     i,j:Integer;
  begin
   { Create FNodes and FLinks from FWNodes and FWLinks }
    if (FWNodes=nil) and (FWLinks=nil) then Exit;
    if (FNodes<>nil) then ClearNodes;
    if (FLinks<>nil) then ClearLinks;
    linkW:=FWLinks;
    nodeW:=FWNodes;
    OldNodeP:=nil;
    OldLinkP:=nil;
    for i:=0 to (FNodesCount-1) do
     begin
       nodeP:=PBPDNode(AllocMem(sizeof(TBPDNode)));
       nodeP^.ssCaption:=nodeW^.ssCaption;
       nodeP^.dwNodeNumber:=nodeW^.dwNodeNumber;
       nodeP^.ssNodeNumber:=nodeW^.ssNodeNumber;
       nodeP^.fABCValue:=NodeW^.fABCValue;
       nodeP^.dwNodeIndex:=NodeW^.dwNodeIndex;
       nodeP^.ssComment:=NodeW^.ssComment;
       for j:=0 to 9 do
       begin
       nodeP^.Left[j]:=NodeW^.Left[j];
       nodeP^.Top[j]:=NodeW^.Top[j];
       nodeP^.AcceptPosition[j]:=NodeW^.AcceptPosition[j];
       end;
       nodeP^.next:=nil;
       if (i=0) then FNodes:=nodeP;
       if (oldNodeP<>nil) then oldNodeP^.next:=nodeP;
       oldNodeP:=nodeP;
       inc(nodeW);
     end;
    for i:=0 to (FLinksCount-1) do
     begin
       linkP:=PBPDLink(AllocMem(sizeof(TBPDLink)));
       linkP^.ssFromCaption:=linkW^.ssFromCaption;
       linkP^.ssToCaption:=linkW^.ssToCaption;
       linkP^.ssGlobalCaption:=linkW^.ssGlobalCaption;
       linkP^.dwFromNodeNumber:=linkW^.dwFromNodeNumber;
       linkP^.dwToNodeNumber:=linkW^.dwToNodeNumber;
       linkP^.dwFlags:=linkW^.dwFlags;
       linkP^.dwFromLinkNumber:=linkW^.dwFromLinkNumber;
       linkP^.dwToLinkNumber:=linkW^.dwToLinkNumber;
       linkP^.dwLinkIndex:=linkW^.dwLinkIndex;
{       linkP^.dwLinkDirectional:=linkW^.dwLinkDirectional;}
       for j:=0 to 9 do
       begin
       linkP^.FromLeft[j]:=linkW^.FromLeft[j];
       linkP^.FromTop[j]:=linkW^.FromTop[j];
       linkP^.FromAcceptPosition[j]:=linkW^.FromAcceptPosition[j];
       linkP^.ToLeft[j]:=linkW^.ToLeft[j];
       linkP^.ToTop[j]:=linkW^.ToTop[j];
       linkP^.ToAcceptPosition[j]:=linkW^.ToAcceptPosition[j];
       end;
       linkP^.DNodeLeft:=linkW^.DNodeLeft;
       linkP^.DNodeTop:=linkW^.DNodeTop;
       linkP^.DNodeAcceptPosition:=linkW^.DNodeAcceptPosition;
       linkP^.ssFromPath:=linkW^.ssFromPath;
       linkP^.ssToPath:=linkW^.ssToPath;
       linkP^.next:=nil;
       if (i=0) then FLinks:=linkP;
       if (oldlinkP<>nil) then oldlinkP^.next:=linkP;
       oldlinkP:=linkP;
       inc(linkW);
     end;
  end;

 procedure   TBPList.Copy;
 Var nodeP:PBPDNode;
     nodeW:PBPNode;
     linkP:PBPDLink;
     linkW:PBPLink;
     i,j:Integer;
  begin
   { Create FWNodes and FWLinks from FNodes and FLinks }
   if (FWNodes<>nil) then FreeMem(FWNodes);
   if (FWLinks<>nil) then FreeMem(FWLinks);
   if (FNodes=nil) and (FLinks=nil) then Exit;
   if (FNodes<>nil) then  FWNodes:=PBPNode(AllocMem(sizeof(TBPNode)*FNodesCount)) else FWNodes:=nil;
   if (FLinks<>nil) then  FWLinks:=PBPLink(AllocMem(sizeof(TBPLink)*FLinksCount)) else FWLinks:=nil;
   linkP:=FLinks;
   nodeP:=FNodes;
   linkW:=FWLinks;
   nodeW:=FWNodes;
   if (FNodes<>nil) then
   begin
    for i:=0 to (FNodesCount-1) do
     begin
      nodeW^.ssCaption:=nodeP^.ssCaption;
      nodeW^.dwNodeNumber:=nodeP^.dwNodeNumber;
      nodeW^.ssNodeNumber:=nodeP^.ssNodeNumber;
      nodeW^.fABCValue:=nodeP^.fABCValue;
      nodeW^.dwNodeIndex:=NodeP^.dwNodeIndex;
      nodeW^.ssComment:=NodeP^.ssComment;
      for j:=0 to 9 do
      begin
      nodeW^.Left[j]:=NodeP^.Left[j];
      nodeW^.Top[j]:=NodeP^.Top[j];
      nodeW^.AcceptPosition[j]:=NodeP^.AcceptPosition[j];
      end;
      inc(nodeW);
      nodeP:=nodeP^.next;
     end;
   end;
   if (FLinks<>nil) then
   begin
    for i:=0 to (FLinksCount-1) do
     begin
      linkW^.ssFromCaption:=linkP^.ssFromCaption;
      linkW^.ssToCaption:=linkP^.ssToCaption;
      linkW^.ssGlobalCaption:=linkP^.ssGlobalCaption;
      linkW^.dwFromNodeNumber:=linkP^.dwFromNodeNumber;
      linkW^.dwToNodeNumber:=linkP^.dwToNodeNumber;
      linkW^.dwFlags:=linkP^.dwFlags;
      linkW^.dwFromLinkNumber:=linkP^.dwFromLinkNumber;
      linkW^.dwToLinkNumber:=linkP^.dwToLinkNumber;
      linkW^.dwLinkIndex:=linkP^.dwLinkIndex;
      for j:=0 to 9 do
      begin
      linkW^.FromLeft[j]:=linkP^.FromLeft[j];
      linkW^.FromTop[j]:=linkP^.FromTop[j];
      linkW^.FromAcceptPosition[j]:=linkP^.FromAcceptPosition[j];
      linkW^.ToLeft[j]:=linkP^.ToLeft[j];
      linkW^.ToTop[j]:=linkP^.ToTop[j];
      linkW^.ToAcceptPosition[j]:=linkP^.ToAcceptPosition[j];
      end;
      linkW^.DNodeLeft:=linkP^.DNodeLeft;
      linkW^.DNodeTop:=linkP^.DNodeTop;
      linkW^.DNodeAcceptPosition:=linkP^.DNodeAcceptPosition;
      linkW^.ssFromPath:=linkP^.ssFromPath;
      linkW^.ssToPath:=linkP^.ssToPath;
{      linkW^.dwLinkDirectional:=linkP^.dwLinkDirectional;}
      inc(linkW);
      linkP:=linkP^.next;
     end;
    end;
  end;

 constructor TBPList.Create(AOwner:TComponent);
  begin
   inherited Create(AOwner);
   FFileStream:=TBPFileStream.Create;
   FNodes:=nil;
   FLinks:=nil;
   FWNodes:=nil;
   FWLinks:=nil;
   FNodesCount:=0;
   FLinksCount:=0;
   FTree:=nil;
   FNTTree:=nil;
   FLTTree:=nil;
   FOpTree:=nil;
{@@@}
   FIPNode:=nil;
{/@@@}
   FEntryArray:=PBPEntryArray(AllocMem(sizeof(TBPEntryArray)));
   FEntryArray^.NodesIndexCount:=0;
   FCheckGraphArray:=PBPCheckGraphArray(AllocMem(sizeof(TBPCheckGraphArray)));
   FCheckGraphArray^.NodesCount:=0;
   FCheckGraphArray^.LinksCount:=0;
   FillChar(FHeader,sizeof(FHeader),0);
   FHeader.dwNodesTypesCount:=0;
   FHeader.dwLinksTypesCount:=0;
   FHeader.dwOpEntries:=0;
   FHeader.dwIPEntries:=0;
   FEditLevel:=False;
   FLevel:=-1;
   FParentLevel:=-2;
   FDebug:=False;
  end;

 destructor  TBPList.Destroy;
  begin
   if (FFileStream<>nil) then FFileStream.Destroy;
   if (FNodes<>nil) or (FLinks<>nil) then ClearCurrent;
   if (FTree<>nil) then ClearAll;
   if (FNTTree<>nil) then ClearNodesTypes;
   if (FLTTree<>nil) then ClearLinksTypes;
   if (FOPTree<>nil) then ClearOperationTree;
{@@@}
   if (FIPNode<>nil) then ClearIPNode;
{/@@@}
   ClearMSDAddon;
   inherited Destroy;
  end;

 procedure   TBPList.SetFileHeader(sModel,sCompany,sAuthor:ShortString;
                                      wDateTime:TDateTime;
                                      wRev:DWORD);
  begin
    FHeader.ssModel:=sModel;
    FHeader.ssCompany:=sCompany;
    FHeader.ssAuthor:=sAuthor;
    FHeader.dfDateTime:=wDateTime;
    FHeader.dwRevision:=wRev;
  end;

 procedure   TBPList.StartNewLevel(wLevel,wParent:Integer;sLevel,sCaption,sNotes:ShortString;
                           wRev:DWORD;wStatus:TDraftStatus);
 Var tmpTree,oldTmpTree:PBPTree;
     i:Integer;
  begin
   if (FEditLevel) then
      begin
        StoreEditingResults(FLevel);
      end;
   tmpTree:=FTree;
   oldTmpTree:=nil;
   while (tmpTree<>nil) do
    begin
     oldTmpTree:=tmpTree;
     tmpTree:=tmpTree^.next;
    end;
    tmpTree:=PBPTree(AllocMem(sizeof(TBPTree)));
    tmpTree^.dwDataOffset:=0;
    tmpTree^.dwLevel:=wLevel;
    tmpTree^.dwParent:=wParent;
    tmpTree^.ssLevel:=sLevel;
    tmpTree^.ssCaption:=sCaption;
    tmpTree^.ssNotes:=sNotes;
    tmpTree^.dwRevision:=wRev;
    tmpTree^.dwNodesCount:=0;
    tmpTree^.dwLinksCount:=0;
    if wStatus=[] then
    tmpTree^.dwStatus:=[dsWorking]
    else
    tmpTree^.dwStatus:=wStatus;
    tmpTree^.dwPageNumber:=0;
    tmpTree^.dwNodesMemRequired:=0;
    tmpTree^.dwLinksMemRequired:=0;
    tmpTree^.lpNodes:=nil;
    tmpTree^.lpLinks:=nil;
    tmpTree^.dwNodesFontSize:=0;
    tmpTree^.dbOverrideNodesFS:=False;
    tmpTree^.dwDNodesFontSize:=0;
    tmpTree^.dbOverrideDNodesFS:=False;
    tmpTree^.dwDProjectionFontSize:=0;
    tmpTree^.dwInProcessFontSize:=0;
    tmpTree^.dbOverrideInProcessFS:=False;
    tmpTree^.dwLabelsFontSize:=0;
    tmpTree^.dbOverrideLabelsFS:=False;
    for i:=0 to 231 do tmpTree^.ssReserved[i]:=0;
    tmpTree^.next:=nil;
    if (oldTmpTree=nil) then FTree:=tmpTree
    else oldTmpTree^.next:=tmpTree;
    FEditLevel:=True;
    FLevel:=wLevel;
    FParentLevel:=wParent;
    ClearCurrent;
  end;

 function    TBPList.SetCurrentLevel(wLevel:Integer):Boolean;
 Var tmpTree:PBPTree;
     flag:Boolean;
  begin
    Result:=False;
    if (FEditLevel) then
      begin
        StoreEditingResults(FLevel);
      end;
    Flag:=True;
    tmpTree:=FTree;
    while (tmpTree<>nil) and (flag) do
     begin
      if (tmpTree^.dwLevel=wLevel) then
       begin
        Flag:=False;
        Result:=True;
        { Some action to set current level }
        ClearCurrent; { Clear FNodes,FLinks,FWNodes,FWLinks }
        FWNodes:=tmpTree^.lpNodes;
        FWLinks:=tmpTree^.lpLinks;
        FNodesCount:=tmpTree^.dwNodesCount;
        FLinksCount:=tmpTree^.dwLinksCount;
        FParentLevel:=tmpTree^.dwParent;
        RecreateNodesAndLinks; { Create FNodes and FLinks from FWNodes and FWLinks}
        FEditLevel:=True; { Set editing level to True }
        FLevel:=wLevel;
       end;
      tmpTree:=tmpTree^.next;
     end;
  end;

 procedure   TBPList.ClearAll; { Clear all tree structure }
 Var treeP,treeP1:PBPTree;
  begin
    if (FEditLevel) then
      begin
        StoreEditingResults(FLevel);
      end;
    if (FTree<>nil) then
     begin
        treeP:=FTree;
        while (treeP<>nil) do
         begin
          treeP1:=treeP;
          treeP:=treeP^.next;
          if (treeP1^.lpNodes<>nil) then FreeMem(treeP1^.lpNodes);
          if (treeP1^.lpLinks<>nil) then FreeMem(treeP1^.lpLinks);
          FreeMem(treeP1);
         end;
     end;
    FTree:=nil;
    FLevel:=-1;
    FEditLevel:=False;
  end;

 { Clear nodes & links lists }
 procedure   TBPList.ClearCurrent;
 Var nodeP,nodeP1:PBPDNode;
     linkP,linkP1:PBPDLink;
  begin
    if (FNodes=nil) and (FLinks=nil) then Exit;
    if (FNodes<>nil) then
      begin
        nodeP:=FNodes;
        while (nodeP<>nil) do
         begin
          nodeP1:=nodeP;
          nodeP:=nodeP^.next;
          FreeMem(nodeP1);
         end;
        FNodesCount:=0;
        FNodes:=nil;
      end;
    if (FLinks<>nil) then
      begin
        linkP:=FLinks;
        while (linkP<>nil) do
         begin
          linkP1:=linkP;
          linkP:=linkP^.next;
          FreeMem(linkP1);
         end;
        FLinksCount:=0;
        FLinks:=nil;
      end;
   if (FWNodes<>nil) then FreeMem(FWNodes);
   if (FWLinks<>nil) then FreeMem(FWLinks);
  end;

 { Add a node }

{@@@}

procedure   TBPList.AddAdvIPNode(NodeCaption,Mech,Upr:ShortString;NodeNumber:DWORD;NodeIndex:DWORD;Comment,Path:ShortString);
Var nodeP,nodeP1:PBPInProcessNode;
               i:Integer;
 begin
   if FIPNode=nil then
     begin
      nodeP:=PBPInProcessNode(AllocMem(sizeof(TBPInProcessNode)));
      nodeP^.dwNodeNumber:=NodeNumber;
      nodeP^.ssCaption:=NodeCaption;
      nodeP^.dwNodeIndex:=NodeIndex;
      nodeP^.ssMech:=Mech;
      nodeP^.ssUpr:=Upr;
      for i:=0 to 9 do
       begin
        nodeP^.Left[i]:=0;
        nodeP^.Top[i]:=0;
        nodeP^.AcceptPosition[i]:=False;
       end;
      nodeP^.ssComment:=Comment;
      nodeP^.ssPath:=Path;
      FHeader.dwIPEntries:=1;
      FIPNode:=nodeP;
      nodeP:=nil;
     end
    else
     begin
      FHeader.dwIPEntries:=FHeader.dwIPEntries+1;
      nodeP:=PBPInProcessNode(AllocMem(sizeof(TBPInProcessNode)*FHeader.dwIPEntries));
      System.Move(FIPNode^,nodeP^,sizeof(TBPInProcessNode)*(FHeader.dwIPEntries-1));
      FreeMem(FIPNode,sizeof(TBPInProcessNode)*(FHeader.dwIPEntries-1));
      nodeP1:=nodeP;
      Inc(nodeP,FHeader.dwIPEntries-1);
      nodeP^.dwNodeNumber:=NodeNumber;
      nodeP^.ssCaption:=NodeCaption;
      nodeP^.dwNodeIndex:=NodeIndex;
      nodeP^.ssMech:=Mech;
      nodeP^.ssUpr:=Upr;
      for i:=0 to 9 do
       begin
        nodeP^.Left[i]:=0;
        nodeP^.Top[i]:=0;
        nodeP^.AcceptPosition[i]:=False;
       end;
      nodeP^.ssComment:=Comment;
      nodeP^.ssPath:=Path;
      FIPNode:=nodeP1;
      nodeP1:=nil;
      nodeP:=nil;
     end;
 end;

procedure   TBPList.AddIPNode(NodeCaption,Mech,Upr:ShortString;NodeNumber:DWORD;NodeIndex:DWORD);
Var nodeP,nodeP1:PBPInProcessNode;
               i:Integer;
 begin
   if FIPNode=nil then
     begin
      nodeP:=PBPInProcessNode(AllocMem(sizeof(TBPInProcessNode)));
      nodeP^.dwNodeNumber:=NodeNumber;
      nodeP^.ssCaption:=NodeCaption;
      nodeP^.dwNodeIndex:=NodeIndex;
      nodeP^.ssMech:=Mech;
      nodeP^.ssUpr:=Upr;
      for i:=0 to 9 do
       begin
        nodeP^.Left[i]:=0;
        nodeP^.Top[i]:=0;
        nodeP^.AcceptPosition[i]:=False;
       end;
      nodeP^.ssComment:='';
      nodeP^.ssPath:='';
      FHeader.dwIPEntries:=1;
      FIPNode:=nodeP;
      nodeP:=nil;
     end
    else
     begin
      FHeader.dwIPEntries:=FHeader.dwIPEntries+1;
      nodeP:=PBPInProcessNode(AllocMem(sizeof(TBPInProcessNode)*FHeader.dwIPEntries));
      System.Move(FIPNode^,nodeP^,sizeof(TBPInProcessNode)*(FHeader.dwIPEntries-1));
      FreeMem(FIPNode,sizeof(TBPInProcessNode)*(FHeader.dwIPEntries-1));
      nodeP1:=nodeP;
      Inc(nodeP,FHeader.dwIPEntries-1);
      nodeP^.dwNodeNumber:=NodeNumber;
      nodeP^.ssCaption:=NodeCaption;
      nodeP^.dwNodeIndex:=NodeIndex;
      nodeP^.ssMech:=Mech;
      nodeP^.ssUpr:=Upr;
      for i:=0 to 9 do
       begin
        nodeP^.Left[i]:=0;
        nodeP^.Top[i]:=0;
        nodeP^.AcceptPosition[i]:=False;
       end;
      nodeP^.ssComment:='';
      nodeP^.ssPath:='';
      FIPNode:=nodeP1;
      nodeP1:=nil;
      nodeP:=nil;
     end;
 end;
{/@@@}



 procedure   TBPList.AddNode(NodeCaption:ShortString;NodeNumber:DWORD;
NodeNumberCaption:TVeryShortString;ABCValue:Double;NodeIndex:DWORD;Comment:ShortString);
 Var nodeP,nodeP1:PBPDNode;
  begin
    if FNodes=nil then
      begin
       nodeP:=PBPDNode(AllocMem(sizeof(TBPDNode)));
       nodeP^.ssCaption:=NodeCaption;
       nodeP^.dwNodeNumber:=NodeNumber;
       nodeP^.ssNodeNumber:=NodeNumberCaption;
       nodeP^.fABCValue:=ABCValue;
       nodeP^.dwNodeIndex:=NodeIndex;
       nodeP^.ssComment:=Comment;
       if nodeP^.ssNodeNumber='' then
          nodeP^.ssNodeNumber:='A'+TVeryShortString(IntToStr(NodeNumber));
       nodeP^.next:=nil;
       FNodes:=nodeP;
       inc(FNodesCount);
      end
    else
      begin
       nodeP1:=FNodes;
       while (nodeP1^.next<>nil) do nodeP1:=nodeP1^.next;
       nodeP:=PBPDNode(AllocMem(sizeof(TBPDNode)));
       nodeP^.ssCaption:=NodeCaption;
       nodeP^.dwNodeNumber:=NodeNumber;
       nodeP^.ssNodeNumber:=NodeNumberCaption;
       nodeP^.fABCValue:=ABCValue;
       nodeP^.dwNodeIndex:=NodeIndex;
       nodeP^.ssComment:=Comment;
       if nodeP^.ssNodeNumber='' then
          nodeP^.ssNodeNumber:='A'+TVeryShortString(IntToStr(NodeNumber));
       nodeP^.next:=nil;
       nodeP1^.next:=nodeP;
       inc(FNodesCount);
      end;
  end;

 { Add a link }

 procedure   TBPList.IntAddLink(LinkFromCaption,LinkToCaption:ShortString;LinkGlobalCaption:TVeryShortString;
                                FromNode,ToNode:DWORD;
                                Flags:DWORD;
                                FromNumber,ToNumber:Integer;
                                LinkIndex:DWORD;
                                FromPath,ToPath:ShortString); { Add Link }
 Var linkP,linkP1:PBPDLink;
     i:Integer;
  begin
   if FLinks=nil then
    begin
      linkP:=PBPDLink(AllocMem(sizeof(TBPDLink)));
      linkP^.ssFromCaption:=LinkFromCaption;
      linkP^.ssToCaption:=LinkToCaption;
      linkP^.ssGlobalCaption:=LinkGlobalCaption;
      linkP^.dwFromNodeNumber:=FromNode;
      linkP^.dwToNodeNumber:=ToNode;
      linkP^.dwFlags:=Flags;
      linkP^.dwLinkIndex:=LinkIndex;
      if ((FromNode and BPGlobalMask)<>0) then
       begin
         linkP^.dwFlags:=linkP^.dwFlags and BP_FROMMASK_INVERT;
         case (FromNode and BPGlobalMask) of
           BPGlobalNodeLeft:  linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_RIGHT;
           BPGlobalNodeRight: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_LEFT;
           BPGlobalNodeTop:   linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_BOTTOM;
           BPGlobalNodeBottom:linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_TOP;
           BPGlobalNodeOperation: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_TOP;
           BPGlobalNodeInProcess: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_RIGHT;
         end;
       end;
      if ((ToNode and BPGlobalMask)<>0) then
       begin
         linkP^.dwFlags:=linkP^.dwFlags and BP_TOMASK_INVERT;
         case (ToNode and BPGlobalMask) of
           BPGlobalNodeLeft:  linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_RIGHT;
           BPGlobalNodeRight: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_LEFT;
           BPGlobalNodeTop:   linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_BOTTOM;
           BPGlobalNodeBottom:linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_TOP;
           BPGlobalNodeOperation: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_TOP;
           BPGlobalNodeInProcess: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_LEFT;
         end;
       end;
      linkP^.dwFromLinkNumber:=FromNumber;
      linkP^.dwToLinkNumber:=ToNumber;
      linkP^.ssFromPath:=FromPath;
      linkP^.ssToPath:=ToPath;
      linkP^.next:=nil;
      FLinks:=linkP;
      inc(FLinksCount);
    end
   else
    begin
      linkP1:=FLinks;
      while (linkP1^.next<>nil) do linkP1:=linkP1^.next;
      linkP:=PBPDLink(AllocMem(sizeof(TBPDLink)));
      linkP^.ssFromCaption:=LinkFromCaption;
      linkP^.ssToCaption:=LinkToCaption;
      linkP^.ssGlobalCaption:=LinkGlobalCaption;
      linkP^.dwFromNodeNumber:=FromNode;
      linkP^.dwToNodeNumber:=ToNode;
      linkP^.dwFlags:=Flags;
      linkP^.dwLinkIndex:=LinkIndex;

      if ((FromNode and BPGlobalMask)<>0) then
       begin
         linkP^.dwFlags:=linkP^.dwFlags and BP_FROMMASK_INVERT;
         case (FromNode and BPGlobalMask) of
           BPGlobalNodeLeft:  linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_RIGHT;
           BPGlobalNodeRight: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_LEFT;
           BPGlobalNodeTop:   linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_BOTTOM;
           BPGlobalNodeBottom:linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_TOP;
           BPGlobalNodeOperation: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_TOP;
           BPGlobalNodeInProcess: linkP^.dwFlags:=linkP^.dwFlags or BP_FROMNODE_RIGHT;
         end;
       end;
      if ((ToNode and BPGlobalMask)<>0) then
       begin
         linkP^.dwFlags:=linkP^.dwFlags and BP_TOMASK_INVERT;
         case (ToNode and BPGlobalMask) of
           BPGlobalNodeLeft:  linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_RIGHT;
           BPGlobalNodeRight: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_LEFT;
           BPGlobalNodeTop:   linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_BOTTOM;
           BPGlobalNodeBottom:linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_TOP;
           BPGlobalNodeOperation: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_TOP;
           BPGlobalNodeInProcess: linkP^.dwFlags:=linkP^.dwFlags or BP_TONODE_LEFT;
         end;
       end;
      linkP^.dwFromLinkNumber:=FromNumber;
      linkP^.dwToLinkNumber:=ToNumber;
      linkP^.ssFromPath:=FromPAth;
      linkP^.ssToPath:=ToPath;
      linkP^.next:=nil;
      linkP1^.next:=linkP;
      inc(FLinksCount);
    end;
  end;

 { Save results to file }
 procedure   TBPList.SaveToFile(FileName:String);
  begin
    {}
    if (FEditLevel) then
      begin
        StoreEditingResults(FLevel);
      end;
    RenameLinksNumbers; { Rename links numbers }
    if FDebug then WriteTree(FTree);
    FFileStream.OpenStream(FileName,stWrite);
    FFileStream.WriteBP(FHeader,FTree,FNTTree,FLTTree,FOPTree,FEntryArray,FCheckGraphArray,FIPNode);
    FFileStream.CloseStream;
  end;

 procedure  TBPList.LoadFromFile(Filename:String);
  begin
    ClearAll;
    ClearNodesTypes;
    ClearLinksTypes;
    ClearMSDAddon;
{@@@}
    ClearIPNode;
{/@@@}
    FTree:=nil;
    FNTTree:=nil;
    FLTTree:=nil;
    FOPTree:=nil;
    FEntryArray:=nil;
    FCheckGraphArray:=nil;
{@@@}
    FIPNode:=nil;
{/@@@}
    FFileStream.OpenStream(FileName,stRead);
    FFileStream.ReadBP(FHeader,FTree,FNTTree,FLTTree,FOPTree,FEntryArray,FCheckGraphArray,FIPNode);
    FFileStream.CloseStream;
  end;

 procedure   TBPList.AddLink(LinkFromCaption,LinkToCaption:ShortString;LinkGlobalCaption:TVeryShortString;
                              FromNode,ToNode:DWORD;
                              FromLinkType,ToLinkType:TLinkType;
                              FromNumber,ToNumber:Integer;LinkIndex:DWORD;USINGFLAGS:DWORD;
                              FromPath,ToPath:ShortString);
  Var Flags:DWORD;
  begin
    Flags:=0;
    case FromLinkType of
     lt_Left:  Flags:=BP_FROMNODE_LEFT;
     lt_Right: Flags:=BP_FROMNODE_RIGHT;
     lt_Top:   Flags:=BP_FROMNODE_TOP;
     lt_Bottom:Flags:=BP_FROMNODE_BOTTOM;
    end;
    case ToLinkType of
     lt_Left:  Flags:=Flags or BP_TONODE_LEFT;
     lt_Right: Flags:=Flags or BP_TONODE_RIGHT;
     lt_Top:   Flags:=Flags or BP_TONODE_TOP;
     lt_Bottom:Flags:=Flags or BP_TONODE_BOTTOM;
    end;
    if Length(LinkFromCaption)>0 then  Flags:=Flags or BP_SHOW_FROMCAPTION;
    if Length(LinkToCaption)>0 then  Flags:=Flags or BP_SHOW_TOCAPTION;
    if Length(LinkGlobalCaption)>0 then  Flags:=Flags or BP_SHOW_GLOBALCAPTION;
{    else
     begin
      if (((FromNode and BPGlobalMask)<>0) or
         ((ToNode and BPGlobalMask)<>0)) and
         (FLevel<>-1) then
          Flags:=Flags or BP_NOTUSED_NOW;
     end;}
    FLAGS:=flags or usingflags;
    IntAddLink(LinkFromCaption,LinkToCaption,LinkGlobalCaption,FromNode,ToNode,
            Flags,FromNumber,ToNumber,LinkIndex,FromPath,ToPath);
  end;

function    TBPList.GetLinkTypeFrom(Flags:DWORD):TLinkType;
  begin
    Result:=lt_Right;
    case (Flags and BP_FROMMASK) of
      BP_FROMNODE_LEFT:   Result:=lt_Left;
      BP_FROMNODE_RIGHT:  Result:=lt_Right;
      BP_FROMNODE_TOP:    Result:=lt_Top;
      BP_FROMNODE_BOTTOM: Result:=lt_Bottom;
    end;
  end;

function    TBPList.GetLinkTypeTo(Flags:DWORD):TLinkType;
  begin
    Result:=lt_Left;
    case (Flags and BP_TOMASK) of
      BP_TONODE_LEFT:   Result:=lt_Left;
      BP_TONODE_RIGHT:  Result:=lt_Right;
      BP_TONODE_TOP:    Result:=lt_Top;
      BP_TONODE_BOTTOM: Result:=lt_Bottom;
    end;
  end;


procedure Register;
begin
  RegisterComponents('DeadSouls', [TBPList]);
end;

end.

