unit TranslateUnit;

interface

Uses Windows,SysUtils,graphics,classes,Forms,RichEdit2,BPDef,BPInt;

Type
     TTokenType=(ttNone,ttFile,ttModel,ttNodeType,ttLinkType,
                 ttLevel,ttNode,ttLink,ttOpNode,ttOpDetail,ttInProcess);

     TBPTranslator=class
       private
        FTranslateStatus: TRichEdit98;
        FSrc:TRichEdit98;
        FBPList:TBPList;
        FFileName:String;
        FWarningCount:Integer;
        FErrorCount:Integer;
        FStrNumber:Integer;
        FNodeTypeIndex:TStringList;
        FLinkTypeIndex:TStringList;
        FLevelNumber:TStringList;
        FNodeNumber:TStringList;
        FLevelStarted:Boolean;
        FCurLevel:Integer;
        FParentLevel:Integer;
        FSSCurLevel:String;
        { OpNode variables }
        FOpDetail:TStringList;
        FOpID:Integer;
        FOpParent:Integer;
        FOpCaption:String;
        FOpNodeIndex:Integer;
        { String routines}
        function AddonFlags(s:String):DWORD;
        function NotCommentLine(Var s:String):Boolean;
        function GetTokenType(Var s:String):TTokenType;
        function GetToken(Var s:String):String;
        function GetBoolean(s:String):Boolean;
        function GetPenStyle(s:String):TPenStyle;
        function GetShadowStyle(s:String):DWORD;
        procedure DeleteQuotes(Var s:String);
        function GetDraftStatus(s:String):TDraftStatus;
        function ConvertGlobals(s:String):Integer;
        function GetColor(s:String;DefColor:TColor):TColor;
        function ConvertSides(s:String;Global:Boolean):TLinkType;
        { Init routines }
        procedure InitTranslateStatus;
        procedure InitBPList;
        { Translate routines }
        procedure TryFileToken(Var s:String);
        procedure TryModelToken(Var s:String);
        procedure TryNodeTypeToken(Var s:String);
        procedure TryLinkTypeToken(Var s:String);
        procedure TryLevelToken(Var s:String);
        procedure TryNodeToken(Var s:String);
        procedure TryLinkToken(Var s:String);
        procedure TryOpNodeToken(Var s:String);
        procedure TryOpDetailToken(Var s:String);
        procedure TryInProcessToken(Var s:String);
        procedure TryAppendOpNode;
        { Warinig and Error procedures }
        procedure AddWarning(msg:String);
        procedure AddError(msg:String);
       public
        constructor Create(Src,Ts:TRichEdit98;BPL:TBPList);
        destructor Destroy; override;
        procedure InitTranslator;
        function  Translate:Boolean;
        procedure SaveResults;
        property FileName:String read FFileName write FFileName;
     end;

implementation



procedure TBPTranslator.AddWarning(msg:String);
Var s:String;
 begin
   Inc(FWarningCount);
   s:=': '+'  '+IntToStr(FStrNumber)+' '+msg;
   FTranslateStatus.Lines.Append(s);
   Application.ProcessMessages;
 end;

procedure TBPTranslator.AddError(msg:String);
Var s:String;
 begin
   Inc(FErrorCount);
   s:=': '+'  '+IntToStr(FStrNumber)+' '+msg;
   FTranslateStatus.Lines.Append(s);
   Application.ProcessMessages;
 end;




procedure TBPTranslator.TryOpNodeToken(Var s:String);
Var s1,s2,s3,s4:String;
 begin
 {}
   s1:=GetToken(s);
   s2:=GetToken(s);
   s3:=GetToken(s);
   DeleteQuotes(s3);
   s4:=GetToken(s);
   FOpID:=StrToIntDef(s1,0);
   FOpParent:=StrToIntDef(s2,0);
   FOpCaption:=s3;
   FOpNodeIndex:=StrToIntDef(s4,0);
   FOpDetail.Clear; { Clear all Op Node Info }
 end;

procedure TBPTranslator.TryOpDetailToken(Var s:String);
Var s1:String;
 begin
 {} s1:=GetToken(s);
    DeleteQuotes(s1);
    FOpDetail.Append(s1);
 end;

procedure TBPTranslator.TryInProcessToken(Var s:String);
Var s1,s2,s3,s4,s5,s6,s7:String;
    nodeNum:Integer;
    nodeInd:Integer;
 begin
 {} s1:=GetToken(s);
    DeleteQuotes(s1);
    s2:=GetToken(s);
    DeleteQuotes(s2);
    s3:=GetToken(s);
    DeleteQuotes(s3);
    s4:=GetToken(s);
    DeleteQuotes(s4);
    s5:=GetToken(s);
    s6:=GetToken(s);
    DeleteQuotes(s6);
    s7:=GetToken(s);
    DeleteQuotes(s7);
    nodeNum:=StrToIntDef(s4,0);
    nodeInd:=StrToIntDef(s5,0);
    FBPList.AddAdvIPNode(s1,s2,s3,nodenum,nodeInd,s6,s7);
 end;

procedure TBPTranslator.TryAppendOpNode;
 begin
   { Add New Op Node }
   FBPList.AddOpNode(FOpID,FOpParent,FOpCaption,FOpNodeIndex,FOpDetail);
 end;

procedure TBPTranslator.TryFileToken(Var s:String);
Var s1:String;
begin
 DeleteQuotes(s);
 FFileName:=s;
 s1:=UpperCase(s);
 if pos('.BPD',s1)=0 then
   begin
     AddError('    ');
   end;
end;

procedure TBPTranslator.TryModelToken(Var s:String);
Var s1,s2,s3,s4:String;
    r:Integer;
begin
  s1:=GetToken(s);
  DeleteQuotes(s1);
  if length(s1)=0 then AddError('  ')
  else
  begin
   s2:=GetToken(s);
   DeleteQuotes(s2);
   s3:=GetToken(s);
   DeleteQuotes(s3);
   s4:=GetToken(s);
   r:=StrToIntDef(s4,1);
   FBPList.SetFileHeader(s1,s2,s3,Now(),r);
  end;
end;

procedure TBPTranslator.TryNodeTypeToken(Var s:String);
Var s1,s2:String;
    i:Integer;
    FontColor,PaintColor,DrawColor,ShadowColor:TColor;
    ShadowStyle:DWORD;
begin
  s1:=GetToken(s);
  i:=StrToIntDef(s1,-1);
  if i=-1 then AddError('   ')
  else
   begin
     if (FNodeTypeIndex.IndexOf(s1)<>-1) then
               AddError('   ')
     else
       begin
         FNodeTypeIndex.Append(s1);
         s2:=GetToken(s);
         DeleteQuotes(s2);
         if length(s2)=0 then AddError('   ')
         else
          begin
           if pos(',',s)=0 then
           FBPList.AddNodeType(i,s2)
           else
            begin
             FontColor:=GetColor(GetToken(s),clBlack);
             PaintColor:=GetColor(GetToken(s),clWhite);
             DrawColor:=GetColor(GetToken(s),clBlack);
             ShadowColor:=GetColor(GetToken(s),clBlack);
             ShadowStyle:=GetShadowStyle(GetToken(s));
             FBPList.AddAdvNodeType(i,s2,FontColor,PaintColor,DrawColor,ShadowColor,ShadowStyle);
            end;
          end;
       end;
   end;
end;

procedure TBPTranslator.TryLinkTypeToken(Var s:String);
Var s1,s2,s3:String;
    i:Integer;
    col:TColor;
    ps:TPenStyle;
begin
  s1:=GetToken(s);
  i:=StrToIntDef(s1,-1);
  if i=-1 then AddError('   ')
  else
   begin
     if (FLinkTypeIndex.IndexOf(s1)<>-1) then
               AddError('   ')
     else
       begin
         FLinkTypeIndex.Append(s1);
         s2:=GetToken(s);
         DeleteQuotes(s2);
         if length(s2)=0 then AddError('   ')
         else
          begin
           if pos(',',s)=0 then
           FBPList.AddLinkType(i,s2)
           else
            begin
              col:=GetColor(GetToken(s),clBlack);
              ps:=GetPenStyle(GetToken(s));
              FBPList.AddAdvLinkType(i,s2,col,ps);
            end;
          end;
       end;
   end;
end;

procedure TBPTranslator.TryLevelToken(Var s:String);
Var s1,s2,s3,s4,s5,s6:String;
    i,j,r:Integer;
    tt:TDraftStatus;
    levNum:Integer;
begin
  s1:=GetToken(s);
  DeleteQuotes(s1); { Level }
  if (FLevelNumber.IndexOfName(s1)<>-1) then
        AddError('  ')
      else
        begin
          s2:=GetToken(s);
          DeleteQuotes(s2); { ParentLevel }
          j:=StrToIntDef(s2,-3);
          if (j=-3) and (FLevelNumber.IndexOfName(s2)=-1)
            then AddError('   ')
          else
           begin
             levNum:=FLevelNumber.Count+1;
             FLevelNumber.Append(s1+'='+IntToStr(levNum));
             if (j=-3) then
                FParentLevel:=StrToIntDef(FLevelNumber.Values[s2],-2)
             else
                if (FLevelNumber.IndexOfName(s2)=-1) then
                  FParentLevel:=j
                else
                  FParentLevel:=StrToIntDef(FLevelNumber.Values[s2],-2);
             s3:=GetToken(s);
             DeleteQuotes(s3);
             if length(s3)=0 then
               AddError('  ')
             else
               begin
                 s4:=GetToken(s);
                 DeleteQuotes(s4);
                 s5:=GetToken(s);
                 r:=StrToIntDef(s5,1);
                 s6:=GetToken(s);
                 tt:=GetDraftStatus(s6);
                 FBPList.StartNewLevel(levNum,FParentLevel,s1,s3,s4,r,tt);
                 FSSCurLevel:=Trim(s1);
                 FLevelStarted:=True;
                 FCurLevel:=levNum;
               end;
           end;
        end;
end;

procedure TBPTranslator.TryNodeToken(Var s:String);
Var s1,s2,s3,s4:String;
    i,j:Integer;
    abc:Double;
    old:Char;
    nodeNum:Integer;
    com:String;
begin
 {}
 if not(FLevelStarted) then Exit;
 s1:=GetToken(s);
 DeleteQuotes(s1);
 if length(s1)=0 then AddError('  ')
 else
  begin
    s2:=GetToken(s);
    DeleteQuotes(s2);
        if (FNodeNumber.IndexOfName(s2)<>-1) then
          AddError('  ')
         else
         begin
          nodeNum:=FNodeNumber.Count+1;
          FNodeNumber.Append(s2+'='+IntToStr(nodeNum));
          s3:=GetToken(s);
          s4:=GetToken(s); {!!!}
          com:=GetToken(s);
          if length(com)>0 then
           DeleteQuotes(com);
          s2:='A'+s2;
          j:=StrToIntDef(s3,-1);
          if j=-1 then AddError('  ')
           else
            begin
              if (FNodeTypeIndex.IndexOf(s3)=-1) then
                AddError('  ')
               else
                begin
                  if length(s4)=0 then s4:='0';
                  old:=DecimalSeparator;
                  DecimalSeparator:='.';
                  abc:=StrToFloat(s4);
                  DecimalSeparator:=old;
                  FBPList.AddNOde(s1,nodeNum,s2,abc,j,com);
                end;
            end;
         end
      end;
end;

procedure TBPTranslator.TryLinkToken(Var s:String);
Var s1,s2,s3,s4,s5,s6,s7,s8,s9:String;
    icam:String;
    flt,tlt:TLinkType;
    LinkFromCaption:String;
    LinkToCaption:String;
    LinkGlobalCaption:String;
    FromNode,ToNode:DWORD;
    MultipleTo:Boolean;
    LinkFromNumber,LinkToNumber:Integer;
    LinkIndex:Integer;
    StartNewLink:Boolean;
    fi,ti:Integer;
    flags:DWORD;
    fromPath,toPath:String;
begin
 { hm hm hm }
 s1:=GetToken(s);
 DeleteQuotes(s1);
 FromNode:=ConvertGlobals(s1);
 if ((FromNode and BPGlobalMask)=0) and
    (FNodeNumber.IndexOfName(s1)=-1) then
     AddError('   ');
 s2:=GetToken(s);
 DeleteQuotes(s2);
 ToNode:=ConvertGlobals(s2);
 if ((ToNode and BPGlobalMask)=0) and
    (FNodeNumber.IndexOfName(s2)=-1) then
     AddError('   ');
 flt:=ConvertSides(GetToken(s),False);
 tlt:=ConvertSides(GetToken(s),False);
 s3:=GetToken(s); { FromCaption }
 DeleteQuotes(s3);
 LinkFromCaption:=s3;
 s4:=GetToken(s);
 DeleteQuotes(s4);
 LinkToCaption:=s4;  { ToCaption }
 LinkGlobalCaption:=GetToken(s);
 DeleteQuotes(LinkGlobalCaption);
 {@@@}
 LinkIndex:=StrToIntDef(GetToken(s),0);
 LinkFromNumber:=StrToIntDef(GetToken(s),1);
 LinkToNumber:=StrToIntDef(GetToken(s),1);
 Flags:=0;
 if ((FromNode and BPGlobalMask)=0) then
   begin
     FromNode:=StrToIntDef(FNodeNumber.Values[s1],0);
     icam:='';
   end else if ((FromNode and BPGlobalNotOpMask)<>0) then
              begin
                icam:='ICAM';
                Flags:=BP_SHOW_GLOBALCAPTION;
              end;
 if ((ToNode and BPGlobalMask)=0) then
  begin
    ToNode:=StrToIntDef(FNodeNumber.Values[s2],0);
    icam:='';
  end else if ((ToNode and BPGlobalNotOpMask)<>0) then
            begin
              icam:='ICAM';
              Flags:=BP_SHOW_GLOBALCAPTION;
            end;
  Flags:=Flags or AddonFlags(GetToken(s));
  FromPath:=GetToken(s);
  DeleteQuotes(FromPath);
  ToPath:=GetToken(s);
  DeleteQuotes(ToPath);
  FBPList.AddLink(LinkFromCaption,LinkToCaption,LinkGlobalCaption,FromNode,ToNode,flt,tlt,
                   LinkFromNumber,LinkToNumber,LinkIndex,Flags,FromPath,ToPath);
end; { Add link }

function TBPTranslator.AddonFlags(s:String):DWORD;
Var s1:String;
 begin
   Result:=0;
   s1:=UpperCase(Trim(s));
   if pos('NOTUSEDFROM',s1)=1 then Result:=BP_NOTUSED_FROM;
   if pos('NOTUSEDTO',s1)=1 then Result:=BP_NOTUSED_TO;
   if pos('NOTUSEDALL',s1)=1 then Result:=BP_NOTUSED_FROM or BP_NOTUSED_TO;
   if pos('NOTUSEDNOW',s1)=1 then Result:=BP_NOTUSED_NOW;
 end;

function TBPTranslator.GetShadowStyle(s:String):DWORD;
Var s1:String;
 begin
   Result:=6;
   s1:=UpperCase(s);
   if pos('NONE',s1)=1 then Result:=0;
   if pos('LEFTBOXED',s1)=1 then Result:=1;
   if pos('LEFTFRAMED',s1)=1 then Result:=2;
   if pos('TOPPERSPECTIVE',s1)=1 then Result:=3;
   if pos('INVPERSPECTIVE',s1)=1 then Result:=4;
   if pos('RIGHTBOXED',s1)=1 then Result:=5;
   if pos('RIGHTFRAMED',s1)=1 then Result:=6;
   if pos('BOTTOMPERSPECTIVE',s1)=1 then Result:=7;
 end;

function TBPTranslator.ConvertGlobals(s:String):Integer;
Var s1:String;
    parnode:Integer;
 begin
  Result:=0;
  s1:=UpperCase(s);
  if (FNodeNumber.IndexOfName(FSSCurLevel)=-1) then
    begin
      parnode:=StrToIntDef(FNodeNumber.Values[FSSCurLevel],0);
    end
  else parnode:=0;
  if pos('GLOBALLEFT',s1)=1 then
             begin
               Result:=BPGlobalLeftNodeNumber or parnode;
             end;
  if pos('GLOBALRIGHT',s1)=1 then
             begin
               Result:=BPGlobalRightNodeNumber or parnode;
             end;
  if pos('GLOBALTOP',s1)=1 then
             begin
               Result:=BPGlobalTopNodeNumber or parnode;
             end;
  if pos('GLOBALBOTTOM',s1)=1 then
             begin
               Result:=BPGlobalBottomNodeNumber or parnode;
             end;
  if pos('GLOBALOPERATION=',s1)=1 then
             begin
               Result:=BPGlobalNodeOperation;
               delete(s1,1,16);
               Result:=Result or StrToIntDef(s1,0);
             end;
  if pos('GLOBALINPROCESS=',s1)=1 then
             begin
               Result:=BPGlobalNodeInProcess;
               delete(s1,1,16);
               Result:=Result or StrToIntDef(s1,0);
             end;
 end;

function TBPTranslator.ConvertSides(s:String;Global:Boolean):TLinkType;
Var s1:String;
 begin
   Result:=lt_Right;
   s1:=UpperCase(s);
   if pos('RIGHTSIDE',s1)=1 then Result:=lt_Right;
   if pos('LEFTSIDE',s1)=1 then Result:=lt_Left;
   if pos('TOPSIDE',s1)=1 then Result:=lt_Top;
   if pos('BOTTOMSIDE',s1)=1 then Result:=lt_Bottom;
 end;

function TBPTranslator.GetDraftStatus(s:String):TDraftStatus;
Var s1:String;
 begin
   Result:=[];
   s1:=UpperCase(s);
   if pos('DRAFT',s1)=1 then Result:=Result+[dsDraft];
   if pos('WORKING',s1)=1 then Result:=Result+[dsWorking];
   if pos('RECOMENDED',s1)=1 then Result:=Result+[dsRecomended];
   if pos('PUBLICATION',s1)=1 then Result:=Result+[dsPublication];
 end;

function TBPTranslator.GetPenStyle(s:String):TPenStyle;
Var s1:String;
 begin
   Result:=psSolid;
   s1:=UpperCase(s);
   if (pos('SOLID',s1)=1) then Result:=psSolid;
   if (pos('DASH',s1)=1) then Result:=psDash;
   if (pos('DOT',s1)=1) then Result:=psDot;
   if (pos('DASHDOT',s1)=1) then Result:=psDashDot;
   if (pos('DASHDOTDOT',s1)=1) then Result:=psDashDotDot;
 end;

function TBPTranslator.GetToken(Var s:String):String;
 begin
   if (pos('"',s)=1) then
    begin
      if (pos('",',s)>0) then
       begin
        Result:=copy(s,1,pos('",',s));
        delete(s,1,pos('",',s)+1);
       end
     else
      begin
        delete(s,1,1);
        if pos('"',s)=0 then
         begin
           Result:=s;
           s:='';
         end
        else
         begin
           Result:='"'+copy(s,1,pos('"',s));
           delete(s,1,pos('"',s));
         end;
      end;
    end
   else
   if pos(',',s)=0 then
    begin
     Result:=s;
     s:='';
    end
   else
    begin
     Result:=copy(s,1,pos(',',s)-1);
     delete(s,1,pos(',',s));
    end;

 end;

procedure TBPTranslator.DeleteQuotes(Var s:String);
 begin
   if pos('"',s)=1 then
   begin
   delete(s,1,1);
   if pos('"',s)=0 then
     AddWarning('  ')
   else
     delete(s,pos('"',s),1);
   end;
 end;

function TBPTranslator.GetColor(s:String;DefColor:TColor):TColor;
Var s1:String;
    i:Integer;
 begin
   Result:=DefColor;
   s1:=UpperCase(s);
   if pos('$',s1)=1 then
     Result:=TColor(StrToInt(s))
   else
     begin
       i:=StrToIntDef(s,-1);
       if i=-1 then
         begin
          if pos('CL',s1)=1 then
            begin
              if pos('CLBLACK',s1)=1 then Result:=clBlack;
              if pos('CLMAROON',s1)=1 then Result:=clMaroon;
              if pos('CLGREEN',s1)=1 then Result:=clGreen;
              if pos('CLOLIVE',s1)=1 then Result:=clOlive;
              if pos('CLNAVY',s1)=1 then Result:=clNavy;
              if pos('CLPURPLE',s1)=1 then Result:=clPurple;
              if pos('CLTEAL',s1)=1 then Result:=clTeal;
              if pos('CLGRAY',s1)=1 then Result:=clGray;
              if pos('CLSILVER',s1)=1 then Result:=clSilver;
              if pos('CLRED',s1)=1 then Result:=clRed;
              if pos('CLLIME',s1)=1 then Result:=clLime;
              if pos('CLYELLOW',s1)=1 then Result:=clYellow;
              if pos('CLBLUE',s1)=1 then Result:=clBlue;
              if pos('CLFUCHSIA',s1)=1 then Result:=clFuchsia;
              if pos('CLAQUA',s1)=1 then Result:=clAqua;
              if pos('CLLTGRAY',s1)=1 then Result:=clLtGray;
              if pos('CLDKGRAY',s1)=1 then Result:=clDkGray;
              if pos('CLWHITE',s1)=1 then Result:=clWhite;
              if pos('CLNONE',s1)=1 then Result:=clNone;
              if pos('CLDEFAULT',s1)=1 then Result:=clDefault;
              if pos('CLSCROLLBAR',s1)=1 then Result:=clScrollBar;
              if pos('CLBACKGROUND',s1)=1 then Result:=clBackground;
              if pos('CLACTIVECAPTION',s1)=1 then Result:=clActiveCaption;
              if pos('CLINACTIVECAPTION',s1)=1 then Result:=clInactiveCaption;
              if pos('CLMENU',s1)=1 then Result:=clMenu;
              if pos('CLWINDOW',s1)=1 then Result:=clWindow;
              if pos('CLWINDOWFRAME',s1)=1 then Result:=clWindowFrame;
              if pos('CLMENUTEXT',s1)=1 then Result:=clMenuText;
              if pos('CLWINDOWTEXT',s1)=1 then Result:=clWindowText;
              if pos('CLCAPTIONTEXT',s1)=1 then Result:=clCaptionText;
              if pos('CLACTIVEBORDER',s1)=1 then Result:=clActiveBorder;
              if pos('CLINACTIVEBORDER',s1)=1 then Result:=clInactiveBorder;
              if pos('CLAPPWORKSPACE',s1)=1 then Result:=clAppWorkSpace;
              if pos('CLHIGHLIGHT',s1)=1 then Result:=clHighlight;
              if pos('CLHIGHLIGHTTEXT',s1)=1 then Result:=clHighlightText;
              if pos('CLBTNFACE',s1)=1 then Result:=clBtnFace;
              if pos('CLBTNSHADOW',s1)=1 then Result:=clBtnShadow;
              if pos('CLGRAYTEXT',s1)=1 then Result:=clGrayText;
              if pos('CLBTNTEXT',s1)=1 then Result:=clBtnText;
              if pos('CLINACTIVECAPTIONTEXT',s1)=1 then Result:=clInactiveCaptionText;
              if pos('CLBTNHIGHLIGHT',s1)=1 then Result:=clBtnHighlight;
              if pos('CL3DDKSHADOW',s1)=1 then Result:=cl3DDkShadow;
              if pos('CL3DLIGHT',s1)=1 then Result:=cl3DLight;
              if pos('CLINFOTEXT',s1)=1 then Result:=clInfoText;
              if pos('CLINFOBK',s1)=1 then Result:=clInfoBk;
            end
          else AddError('  ');
         end
       else
         Result:=TColor(StrToInt(s));
     end;
 end;

function TBPTranslator.GetBoolean(s:String):Boolean;
Var s1:String;
 begin
   Result:=False;
   s1:=UpperCase(s);
   if pos('TRUE',s1)=1 then Result:=True;
   if pos('FALSE',s1)=1 then Result:=False;
 end;

function TBPTranslator.NotCommentLine(Var s:String):Boolean;
 begin
   Result:=True;
   s:=Trim(s);
  { s:=UpperCase(s);}
   if pos(';',s)=1 then Result:=False;
   if pos('{',s)=1 then Result:=False;
   if pos('/*',s)=1 then Result:=False;
 end;

function TBPTranslator.GetTokenType(Var s:String):TTokenType;
Var s1:String;
 begin
   Result:=ttNone;
   s1:=UpperCase(s);
   if pos('FILE,',s1)=1 then Result:=ttFile;
   if pos('MODEL,',s1)=1 then Result:=ttModel;
   if pos('NODETYPE,',s1)=1 then Result:=ttNodeType;
   if pos('LINKTYPE,',s1)=1 then Result:=ttLinkType;
   if pos('LEVEL,',s1)=1 then Result:=ttLevel;
   if pos('NODE,',s1)=1 then Result:=ttNode;
   if pos('LINK,',s1)=1 then Result:=ttLink;
   if pos('OPNODE,',s1)=1 then Result:=ttOpNode;
   if pos('OPDETAIL,',s1)=1 then Result:=ttOpDetail;
   if pos('INPROCESSNODE,',s1)=1 then Result:=ttInProcess;
   case Result of
     ttNone: begin end;
     ttFile:        delete(s,1,5);
     ttModel:       delete(s,1,6);
     ttNodeType:    delete(s,1,9);
     ttLinkType:    delete(s,1,9);
     ttLevel:       delete(s,1,6);
     ttNode:        delete(s,1,5);
     ttLink:        delete(s,1,5);
     ttOpNode:      delete(s,1,7);
     ttOpDetail:    delete(s,1,9);
     ttInProcess:   delete(s,1,14);
   end;
 end;

function  TBPTranslator.Translate:Boolean;
var i,cnt:Integer;
    s:String;
    tt:TTokenType;
    startOpNode:Boolean;
 begin
   startOpNode:=False;
   cnt:=FSrc.Lines.Count-1;
   for i:=0 to cnt do
    begin
      s:=FSrc.Lines.Strings[i];
      FStrNumber:=i;
      if NotCommentLine(s) then
        begin
          tt:=GetTokenType(s);
          case (tt) of
            ttNone:          begin end;
            ttFile:          TryFileToken(s);
            ttModel:         TryModelToken(s);
            ttNodeType:      TryNodeTypeToken(s);
            ttLinkType:      TryLinkTypeToken(s);
            ttLevel:         begin
                              if startOpNode then
                                begin
                                  TryAppendOpNode;
                                  startOpNode:=False;
                                end;
                              TryLevelToken(s);
                             end;
            ttNode:          TryNodeToken(s);
            ttLink:          TryLinkToken(s);
            ttOpNode:        begin
                              if startOpNode then TryAppendOpNode;
                              startOpNode:=True;
                              TryOpNodeToken(s);
                             end;
            ttOpDetail:      TryOpDetailToken(s);
            ttInProcess:     begin
                              if startOpNode then
                                begin
                                  TryAppendOpNode;
                                  startOpNode:=False;
                                end;
                              TryInProcessToken(s);
                             end;
          end;
        end;
    end;
  Result:=(FErrorCount=0) and (FWarningCount=0);
 end;

procedure TBPTranslator.SaveResults;
 begin
   FBPList.SaveToFile(FFileName);
 end;

procedure TBPTranslator.InitTranslateStatus;
 begin
   FTranslateStatus.Clear;
 end;

procedure TBPTranslator.InitBPList;
 begin
   { Clear whole BPList }
   FBPList.ClearAll;
   FBPList.ClearLinksTypes;
   FBPList.ClearNodesTypes;
   FBPList.ClearOperationTree;
   FBPList.ClearIPNode;
 end;

procedure TBPTranslator.InitTranslator;
 begin
   InitTranslateStatus;
   InitBPList;
   FFileName:='';
   FWarningCount:=0;
   FErrorCount:=0;
   FStrNumber:=0;
   FLevelStarted:=False;
   FNodeTypeIndex.Clear;
   FLinkTypeIndex.Clear;
   FLevelNumber.Clear;
   FNodeNumber.Clear;
 end;

constructor TBPTranslator.Create(Src,Ts:TRichEdit98;BPL:TBPList);
 begin
   inherited Create;
   FNodeTypeIndex:=TStringList.Create;
   FLinkTypeIndex:=TStringList.Create;
   FLevelNumber:=TStringList.Create;
   FNodeNumber:=TStringList.Create;
   FOpDetail:=TStringList.Create;
   FTranslateStatus:=Ts;
   FSrc:=Src;
   FBPList:=BPL;
 end;

destructor TBPTranslator.Destroy;
 begin
   FNodeTypeIndex.Free;
   FLinkTypeIndex.Free;
   FLevelNumber.Free;
   FNodeNumber.Free;
   FOpDetail.Free;
   FTranslateStatus:=nil;
   FSrc:=nil;
   FBPList:=nil;
   inherited Destroy;
 end;

end.
