Unit HaspUnit;

interface

Uses WinTypes, WinProcs;

Type  TCustomMemoHasp=class(TPersistent)
      end;


Const
(*
 * A list of the HASP services
 *)
   IS_HASP                      =  1;
   GET_HASP_CODE                =  2;
   READ_MEMO                    =  3;
   WRITE_MEMO                   =  4;
   GET_HASP_STATUS              =  5;
   GET_ID_NUM                   =  6;
   READ_MEMO_BLOCK              = 50;
   WRITE_MEMO_BLOCK             = 51;

   MEMO_INCORRECT_PASSWORD      = -3;

   LPT_IBM_ALL_HASP25  = 0;
   LPT_IBM_ALL_HASP36  = 50;
   LPT_NEC_ALL_HASP36  = 60;

   MEMO_BUFFER_SIZE  = 248;
   TIME_BUFFER_SIZE  = 16; (* TimeHASP maximum block size *)

   (*
 * A list of the TimeHASP services
 *)
   TIMEHASP_SET_TIME            = 70;
   TIMEHASP_GET_TIME            = 71;
   TIMEHASP_SET_DATE            = 72;
   TIMEHASP_GET_DATE            = 73;
   TIMEHASP_WRITE_MEMORY        = 74;
   TIMEHASP_READ_MEMORY         = 75;
   TIMEHASP_WRITE_MEMORY_BLOCK  = 76;
   TIMEHASP_READ_MEMORY_BLOCK   = 77;
   TIMEHASP_GET_ID_NUM          = 78;

   TIME_INCORRECT_PASSWORD      =-28;

(*
 * A list of NetHASP services.
 *)
   NET_LAST_STATUS              = 40;
   NET_GET_HASP_CODE            = 41;
   NET_LOGIN                    = 42;
   NET_LOGOUT                   = 43;
   NET_READ_WORD                = 44;
   NET_WRITE_WORD               = 45;
   NET_GET_ID_NUMBER            = 46;
   NET_SET_IDLE_TIME            = 48;
   NET_READ_MEMO_BLOCK          = 52;
   NET_WRITE_MEMO_BLOCK         = 53;

   OK                           = 0;
   NET_READ_ERROR               = 131;
   NET_WRITE_ERROR              = 132;


(*
 * Global variables.
 *)

Var
  LptNum ,SeedCode : longint;
  Pass1            : longint;
  Pass2            : longint;
  p1,p2,p3,p4      : longint;
  Service          : longint;
  ID               : longint;
  i                : integer;
  HaspMemoBuff     : array[1..MEMO_BUFFER_SIZE] of word ;
  TimeHaspMemoBuff : array[1..TIME_BUFFER_SIZE] of byte ;
  ProductType      : integer;


Function  Is_MemoHASP :Boolean;
function  Check_Hasp:Boolean;
function  GetID:Boolean;

implementation

procedure  hasp (Service, SeedCode, LptNum, Pass1, Pass2 : longint;
                     var p1,p2,p3,p4 : longint);external;


{$L  hasp32b.obj}

Function Is_MemoHASP :Boolean;
begin
    Clear_Cnts;
    Set_PassWords;
    Is_MemoHASP := True;
    Service := GET_HASP_STATUS;
    hasp(Service, SeedCode, LPT_IBM_ALL_HASP25, Pass1, Pass2, p1, p2, p3, p4);
    if (p2 = 3) or ((p2 = 0) and (p3 <> 0)) Then
    begin
        dlg_MemoHASPTest.ed_OutMsg.Text := 'Not a MemoHASP.';
        Is_MemoHASP := False;
    end
end;

function  Check_Hasp:Boolean;
begin
    Check_Hasp := True;
    Clear_Cnts;
    Service := IS_HASP;
    hasp ( Service, SeedCode, LPT_IBM_ALL_HASP25, Pass1, Pass2, p1, p2, p3, p4 );
    if ( p1 = 0 ) then
    begin
        dlg_MemoHASPTest.ed_OutMsg.Text :='No HASP key was found.';
        Check_Hasp := False;
    end
end;

(*
 * Read the ID number.
 * The ID number is a 32 bit integer which is calculated as follows:
 * ID = p1 + 65536 * p2
 *)
function GetID:Boolean;
begin
    Result:=False;
    if Check_Hasp = False then exit;
    ed_IDNum.Text := '';
    if Is_MemoHASP = False then exit;
    p1 := 0;
    p2 := 0;
    Service := GET_ID_NUM;
    hasp ( Service, SeedCode, LptNum, Pass1, Pass2, p1, p2, p3, p4 );
     { Check the status }
    if p3 = MEMO_INCORRECT_PASSWORD then
        ed_OutMsg.Text := 'Incorrect Password.'
    else
      if p3 <> 0 then
        begin
          ed_OutMsg.Text := 'Get ID Number Failed. Error code = ' + IntToStr(p3);
        end
      else
        begin
          ID := word(p2);
          ID := ID shl 16;
          ID := ID + word(p1);
          ed_OutMsg.Text := 'Get ID Number OK.';
          ed_IDNum.Text := IntToStr(ID);
       end;
end;


end.