{$INCLUDE VERSIONS.INC}

unit Menu97Edit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Menus, Menus97_V3, Grids, Buttons, DsgnIntf, ComCtrls;

type
  TMenu97Editor = class(TForm)
    Panel1: TPanel;
    MenuView: TTreeView;
    Panel2: TPanel;
    AddItem: TButton;
    AddSubItem: TButton;
    RemoveItem: TButton;
    PopupMenu97_V31: TPopupMenu97_V3;
    miAddItem: TMenuItem97;
    miRemoveItem: TMenuItem97;
    miAddSubitem: TMenuItem97;
    miSep1: TMenuItem97;
    miInsertFromResource: TMenuItem97;
    ImageList: TImageList;
    miSep2: TMenuItem97;
    miSaveTemplate: TMenuItem97;
    miInsertFromTemplate: TMenuItem97;
    miDeleteTemplate: TMenuItem97;
    Panel3: TPanel;
    PreviewPanel: TPaintBox;
    procedure MenuViewChange(Sender: TObject; Node: TTreeNode);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure AddItemClick(Sender: TObject);
    procedure RemoveItemClick(Sender: TObject);
    procedure MenuViewKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure AddSubItemClick(Sender: TObject);
    procedure MenuViewEdited(Sender: TObject; Node: TTreeNode;
      var S: String);
    procedure MenuViewDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure MenuViewDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure MenuViewDblClick(Sender: TObject);
    procedure PreviewPanelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Panel3MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    FMenu: TMenu;
    FOptions: TMenus97Options;
    FDesigner: IFormDesigner;
    FMenuItem97: TMenuItem97;
    FStop: Boolean;
    FMouseIn: Boolean;
    procedure SetMenu(Value: TMenu);
    procedure SetOptions(Value: TMenus97Options);
    {$ifdef D4_OR_GREATER}
    procedure MenuChanged(Sender: TObject; Source: TMenuItem; Rebuild: Boolean);
    {$else}
    procedure MenuChanged(Sender: TObject; Rebuild: Boolean);
    {$endif}
    procedure DrawPreview(Item: TMenuItem97; Selected: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DestroyWnd; override;
    property Menu: TMenu read FMenu write SetMenu;
    property Options: TMenus97Options read FOptions write SetOptions;
    property Designer: IFormDesigner read FDesigner write FDesigner;
  end;

Var MenuEditorList: TList;

implementation

{$R *.DFM}

constructor TMenu97Editor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  MenuEditorList.Add(Self);
  FStop:= False;
  FMouseIn:= False;
end;

destructor TMenu97Editor.Destroy;
begin
  MenuEditorList.Remove(Self);
  inherited Destroy;
end;

procedure TMenu97Editor.DestroyWnd;
begin
  inherited DestroyWnd;
  Self:= nil;
end;

procedure TMenu97Editor.DrawPreview(Item: TMenuItem97; Selected: Boolean);
begin
  if ( FStop ) then Exit;
  PreviewPanel.Canvas.FillRect(PreviewPanel.ClientRect);
  Options.Canvas.Handle:= PreviewPanel.Canvas.Handle;
  if Selected then
    Options.DrawItem(Item, PreviewPanel.ClientRect, [odSelected])
  else
    Options.DrawItem(Item, PreviewPanel.ClientRect, [])
end;

{$ifdef D4_OR_GREATER}
procedure TMenu97Editor.MenuChanged(Sender: TObject; Source: TMenuItem; Rebuild: Boolean);
{$else}
procedure TMenu97Editor.MenuChanged(Sender: TObject; Rebuild: Boolean);
{$endif}
begin
  if ( FStop ) then Exit;
  if ( MenuView.Selected <> nil ) and ( MenuView.Selected.Data <> nil ) then begin
    MenuView.Selected.Text:= TMenuItem97(MenuView.Selected.Data).Caption;
    MenuView.Selected.ImageIndex:= TMenuItem97(MenuView.Selected.Data).ImageIndex;
    MenuView.Selected.SelectedIndex:= TMenuItem97(MenuView.Selected.Data).ImageIndex;
    Options.ModifyMenuTree(TMenuItem97(Menu.Items), False);
    DrawMenuBar(Options.FormHandle);
    DrawPreview(TMenuItem97(MenuView.Selected.Data), False);
  end;
end;

procedure TMenu97Editor.SetOptions(Value: TMenus97Options);
begin
  if ( Value <> nil ) then
    FOptions:= Value;
end;

procedure TMenu97Editor.SetMenu(Value: TMenu);
Var i: Integer;

  procedure AddMenuItem (MenuItem : TMenuItem97);
  var j        : integer;
      TreeNode : TTreeNode;
      Node     : TTreeNode;
  begin
    TreeNode := MenuView.Items.AddChildObject (MenuView.Selected,MenuItem.Caption, Pointer(MenuItem));
    TreeNode.ImageIndex:=MenuItem.ImageIndex;
    TreeNode.SelectedIndex:=MenuItem.ImageIndex;
    for j := 0 to MenuItem.Count - 1 do begin
      MenuView.Selected := TreeNode;
      Node := MenuView.Items.AddChildObject(MenuView.Selected, MenuItem.Items [j].Caption, Pointer(MenuItem.Items[j]));
      Node.ImageIndex:= MenuItem[j].ImageIndex;
      Node.SelectedIndex:= MenuItem[j].ImageIndex;
      if MenuItem.Items[j].Count > 0 then
        AddMenuItem (MenuItem.Items [j]);
    end;
  end;

begin
  if ( Value <> nil ) then begin
    FMenu:= Value;
    MenuView.Items.Clear;
    if ( Options <> nil ) then
      MenuView.Images:= Options.Images;
    for i := 0 to Menu.Items.Count - 1 do begin
      MenuView.Selected := nil;
      AddMenuItem (TMenuItem97(Menu.Items [i]));
    end;
  end;
end;

procedure TMenu97Editor.MenuViewChange(Sender: TObject; Node: TTreeNode);
begin
  if ( FStop ) then Exit;
  if ( Node <> nil ) and ( Node.Data <> nil ) then begin
    Designer.SelectComponent(TMenuItem97(Node.Data));
    PreviewPanel.Height:= Options.DefaultHeight;
    DrawPreview(TMenuItem97(Node.Data), False);
    TMenuItem97(Node.Data).OnChange:= MenuChanged;
  end;
end;

procedure TMenu97Editor.MenuViewEdited(Sender: TObject; Node: TTreeNode;
  var S: String);
begin
  if ( FStop ) then Exit;
  TMenuItem97(MenuView.Selected.Data).Caption:= S;
  Options.ModifyMenuTree(TMenuItem97(Menu.Items), False);
  DrawMenuBar(Options.FormHandle);
end;

procedure TMenu97Editor.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  FStop:= True;
  Action:= caFree;
  MenuView.OnChange:= nil;
  if ( MenuView.Selected <> nil )  and ( MenuView.Selected.Data <> nil ) then
    TMenuItem97(MenuView.Selected.Data).OnChange:= nil;
  MenuView.Items.Clear;
end;

procedure TMenu97Editor.AddItemClick(Sender: TObject);
Var Index: Integer;
    TreeNode: TTreeNode;
    Node    : TTreeNode;
begin
  FMenuItem97:= NewItem97('MenuItem97', 0, False, True, nil, 0, Designer.UniqueName('MenuItem97'));
  Index:= -1;
  TreeNode:= nil;
  if ( MenuView.Selected <> nil ) then begin
    Index:= MenuView.Selected.Index;
    if( MenuView.Selected.Level <> 0 ) then
      TreeNode:= MenuView.Selected.Parent;
  end;
  if ( TreeNode = nil ) then begin
    if ( FMenu is TMainMenu97_V3 ) then
      FMenuItem97.Look:= il3D;
    InsertMenuItems97(Menu, Index, FMenuItem97);
  end
  else begin
    InsertItems97(Menu, TMenuItem97(TreeNode.Data), Index, FMenuItem97);
  end;
  if ( Index <> -1 ) then begin
    Node := MenuView.Items.InsertObject (MenuView.Selected, FMenuItem97.Name, Pointer(FMenuItem97));
  end
  else
    Node := MenuView.Items.AddChildObject (TreeNode, FMenuItem97.Caption, Pointer(FMenuItem97));
  Node.ImageIndex:= FMenuItem97.ImageIndex;
  Node.SelectedIndex:= FMenuItem97.ImageIndex;
  MenuView.Selected:= Node;
  Options.ModifyMenuTree(TMenuItem97(Menu.Items), False);
  DrawMenuBar(Options.FormHandle);
end;

procedure TMenu97Editor.RemoveItemClick(Sender: TObject);
begin
  if ( MenuView.Selected <> nil ) and ( MenuView.Selected.Data <> nil ) then begin
    Options.ModifyMenuTree(TMenuItem97(MenuView.Selected.Data), True);
    if ( TMenuItem97(MenuView.Selected.Data).Parent <> nil ) and ( TMenuItem97(MenuView.Selected.Data).Parent is TMenuItem97 ) then begin
      TMenuItem97(MenuView.Selected.Data).OnChange:= Nil;
      //TMenuItem97(MenuView.Selected.Data).Parent.Remove(TMenuItem97(MenuView.Selected.Data));
      //TMenuItem97(MenuView.Selected.Data).Delete(TMenuItem97(MenuView.Selected.Data).MenuIndex);
      TMenuItem97(MenuView.Selected.Data).Parent.Delete(TMenuItem97(MenuView.Selected.Data).MenuIndex);
      TMenuItem97(MenuView.Selected.Data).Free;
      MenuView.Selected.Delete;
      MenuView.Selected:= nil;
    end
    else begin
      TMenuItem97(MenuView.Selected.Data).OnChange:= Nil;
      //FMenu.Items.Remove(TMenuItem97(MenuView.Selected.Data));
      //TMenuItem97(MenuView.Selected.Data).Delete(TMenuItem97(MenuView.Selected.Data).MenuIndex);
      FMenu.Items.Delete(TMenuItem97(MenuView.Selected.Data).MenuIndex);
      TMenuItem97(MenuView.Selected.Data).Free;
      MenuView.Items.Delete(MenuView.Selected);
      MenuView.Selected:= nil;
    end;
  end;
  Options.ModifyMenuTree(TMenuItem97(Menu.Items), False);
  DrawMenuBar(Options.FormHandle);
end;

procedure TMenu97Editor.AddSubItemClick(Sender: TObject);
Var TreeNode: TTreeNode;
    Node    : TTreeNode;
begin
  if ( MenuView.Selected <> nil ) then begin
    TreeNode:= MenuView.Selected;
    FMenuItem97:= NewItem97('MenuItem97', 0, False, True, nil, 0, Designer.UniqueName('MenuItem97'));
    Menu.Owner.InsertComponent(FMenuItem97);
    TMenuItem97(TreeNode.Data).Add(FMenuItem97);
    Node:= MenuView.Items.AddChildObject (TreeNode, FMenuItem97.Caption, Pointer(FMenuItem97));
    Node.ImageIndex:= FMenuItem97.ImageIndex;
    Node.SelectedIndex:= FMenuItem97.ImageIndex;
    MenuView.Selected:= Node;
    FMenuItem97.OnChange:= MenuChanged;
    Options.ModifyMenuTree(TMenuItem97(Menu.Items), False);
    DrawMenuBar(Options.FormHandle);
  end;
end;

procedure TMenu97Editor.MenuViewKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_DELETE : begin
                  RemoveItemClick(Self);
                end;
    VK_INSERT : begin
                  if ( ssCtrl in Shift ) then
                    AddSubItemClick(Self)
                  else
                    AddItemClick(Self);
                end;
  end;
end;

procedure TMenu97Editor.MenuViewDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept:= ( Source is TTreeView ) and ( Source = MenuView ) and
           ( ( MenuView.GetNodeAt(X, Y) <> nil ) and
           ( MenuView.GetNodeAt(X, Y) <> (Source as TTreeView).Selected ) and
           Not ( MenuView.GetNodeAt(X, Y).HasAsParent((Source as TTreeView).Selected) ) ) or
           ( MenuView.GetNodeAt(X, Y) = nil );
end;

procedure TMenu97Editor.MenuViewDragDrop(Sender, Source: TObject; X,
  Y: Integer);
Var TreeNodeDrop: TTreeNode;
    TreeNodeDrag: TTreeNode;
    Node        : TTreeNode;

  procedure AddChilds (OldParent, NewParent: TTreeNode);
  Var i: Integer;
  begin
    for i:= 0 to OldParent.Count - 1 do begin
      FMenuItem97:= OldParent.Item[i].Data;
      Node:= MenuView.Items.AddChildObject (NewParent, FMenuItem97.Caption, Pointer(FMenuItem97));
      Node.ImageIndex:= FMenuItem97.ImageIndex;
      Node.SelectedIndex:= FMenuItem97.ImageIndex;
      AddChilds (OldParent.Item[i], Node);
    end;
  end;

begin
  FMenuItem97:= TMenuItem97((Source as TTreeView).Selected.Data);
  TreeNodeDrag:= MenuView.Selected;
  FMenuItem97:= TMenuItem97(TreeNodeDrag.Data);
  TreeNodeDrop:= MenuView.GetNodeAt(X, Y);
  if ( TreeNodeDrop = nil ) then begin
    Node:= MenuView.Items.AddObject (nil, FMenuItem97.Caption, Pointer(FMenuItem97));
    FMenuItem97.MenuIndex:= MenuView.Items.Count - 1;
  end
  else begin
    Node:= MenuView.Items.InsertObject (TreeNodeDrop, FMenuItem97.Caption, Pointer(FMenuItem97));
    FMenuItem97.SetParentComponent(TMenuItem97(TreeNodeDrop.Data).Parent);
    FMenuItem97.MenuIndex:= TreeNodeDrop.Index - 1;
  end;
  Node.ImageIndex:= FMenuItem97.ImageIndex;
  Node.SelectedIndex:= FMenuItem97.ImageIndex;
  AddChilds (TreeNodeDrag, Node);
  MenuView.Selected:= Node;
  FMenuItem97.OnChange:= MenuChanged;
  MenuView.Items.Delete(TreeNodeDrag);
end;

procedure TMenu97Editor.MenuViewDblClick(Sender: TObject);
begin
  if ( MenuView.Selected = nil ) then Exit;
  GetComponentEditor(TMenuItem97(MenuView.Selected.data), Designer).Edit;
end;

procedure TMenu97Editor.PreviewPanelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if ( FStop ) then Exit;
  if PtInRect(PreviewPanel.ClientRect, Point(X, Y)) then begin
    if ( Panel3.Handle <> GetCapture ) then
      SetCapture(Panel3.Handle);
    if ( Not FMouseIn ) then
      FMouseIn:= True
    else
      Exit;
  end
  else begin
    if ( Not FMouseIn ) then
      Exit;
    FMouseIn:= False;
    ReleaseCapture;
  end;
  if ( MenuView.Selected <> nil ) then begin
    DrawPreview( TMenuItem97(MenuView.Selected.Data), FMouseIn);

  end;
end;

procedure TMenu97Editor.Panel3MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if ( FStop ) then Exit;
  PreviewPanelMouseMove(Sender, Shift, X, Y);
end;

initialization
  MenuEditorList:= TList.Create;
finalization
  while MenuEditorList.Count > 0 do
    TMenu97Editor(MenuEditorList[0]).Free;
  MenuEditorList.Free;
end.
