unit UnderlineComboBox;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,RichEdit2;

type
  TUnderlineComboBox = class(TCustomComboBox)
  private
    FUnderlineValue:TUnderlineType;
    FOnChange: TNotifyEvent;
    procedure SetUnderlineValue(NewValue: TUnderlineType);
    procedure ResetItemHeight;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    { Private declarations }
  protected
    { Protected declarations }
    procedure CreateWnd; override;
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure Click; override;
    procedure BuildList; virtual;
    procedure Change; dynamic;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    property Text;
  published
    { Published declarations }
    property UnderlineValue: TUnderlineType read FUnderlineValue write SetUnderlineValue default ultNone;
    property Color;
    property Ctl3D;
    property DragMode;
    property DragCursor;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
  end;

Const UnderlinesInList=8;
      UnderlineValues: array [1..UnderlinesInList] of TUnderlineType =
      (ultNone,ultSingle, ultDotted, ultWave,
                    ultThick, ultHair, ultDashDD, ultDashD);
      UnderlineNames: array [1..UnderlinesInList] of String =
      ('()',' ',' ',' ',' ',' ',' ',' ');
procedure Register;

implementation

{ Utility routine }

function GetItemHeight(Font: TFont): Integer;
var
  DC: HDC;
  SaveFont: HFont;
  Metrics: TTextMetric;
begin
  DC := GetDC(0);
  try
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextMetrics(DC, Metrics);
    SelectObject(DC, SaveFont);
  finally
    ReleaseDC(0, DC);
  end;
  Result := Metrics.tmHeight + 1;
end;



constructor TUnderlineComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Style := csOwnerDrawFixed;
  FUnderlineValue := ultNone;  { make default color selected }
end;

procedure TUnderlineComboBox.BuildList;
var
  I: Integer;
  UnderlineName: string;
begin
  Clear;
  for I := 1 to UnderlinesInList do begin
    { delete two first characters which prefix "cl" educated }
    UnderlineName := UnderlineNames[i];
    Items.AddObject(UnderlineName, TObject(UnderlineValues[I]));
  end;
end;

procedure TUnderlineComboBox.SetUnderlineValue(NewValue: TUnderlineType);
var
  Item: Integer;
  CurrentUnderline: TUnderlineType;
begin
  if (ItemIndex < 0) or (NewValue <> FUnderlineValue) then
    { change selected item }
    for Item := 0 to Pred(Items.Count) do begin
      CurrentUnderline := TUnderlineType(Items.Objects[Item]);
      if CurrentUnderline = NewValue then begin
        FUnderlineValue := NewValue;
        if ItemIndex <> Item then ItemIndex := Item;
        Change;
        Break;
      end;
    end;
end;

procedure TUnderlineComboBox.CreateWnd;
begin
  inherited CreateWnd;
  BuildList;
  SetUnderlineValue(FUnderlineValue);
end;

procedure TUnderlineComboBox.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  ARect: TRect;
  Text: array[0..255] of Char;
  Safer: TPenStyle;
  SColor:TColor;
  CurrentUnderline:TUnderlineType;
  i,hh,ww:Integer;
begin
  ARect := Rect;
  Inc(ARect.Top, 2);
  Inc(ARect.Left, 2);
  Dec(ARect.Bottom, 2);
  Dec(ARect.Right, 3);
  hh:=(ARect.Top+ARect.Bottom) div 2;
  ww:=(ARect.Right-ARect.Left) div 6;
  with Canvas do begin
    FillRect(Rect);
    Safer:=Pen.Style;
    SColor:=Pen.Color;
    Pen.Color := clBlack;
    CurrentUnderline:=TUnderLineType(Items.Objects[Index]);
    try
     case CurrentUnderline of
      ultNone: begin end;
      ultSingle: begin
                    Pen.Style:=psSolid;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);
                 end;
      ultDotted: begin
                    Pen.Style:=psDot;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);
                 end;
      ultWave:   begin { Dash }
                    Pen.Style:=psDash;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);
                 end;
      ultThick:  begin { Dash Dot }
                    Pen.Style:=psDashDot;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);

                 end;
      ultHair:   begin { Dash Dot Dot }
                    Pen.Style:=psDashDotDot;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);
                 end;
      ultDashDD: begin { Wave }
                    Pen.Style:=psSolid;
                    for i:=0 to (ww-1) do
                     begin
                      MoveTo(ARect.Left+i*6,hh-1);
                      LineTo(ARect.Left+i*6+3,hh+1);
                      MoveTo(ARect.Left+i*6+3,hh+1);
                      LineTo(ARect.Left+(i+1)*6,hh-1);
                     end;
                  end;
      ultDashD:  begin { Thick}
                    Pen.Style:=psSolid;
                    MoveTo(Arect.Left,hh);
                    LineTo(ARect.Right,hh);
                    MoveTo(Arect.Left,hh+1);
                    LineTo(ARect.Right,hh+1);
                 end;
     end;
    finally
      Pen.Style:=Safer;
      Pen.Color:=SColor;
    end;
    if (Index<1) then begin
      StrPCopy(Text, Items[Index]);
      Rect.Left := Rect.Left + 6;
      DrawText(Canvas.Handle, Text, StrLen(Text), Rect, DT_SINGLELINE or
        DT_VCENTER or DT_NOPREFIX);
    end;
  end;
end;

procedure TUnderlineComboBox.Click;
begin
  if ItemIndex >= 0 then UnderlineValue := TUnderlineType(Items.Objects[ItemIndex]);
  inherited Click;
end;

procedure TUnderlineComboBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  ResetItemHeight;
  RecreateWnd;
end;

procedure TUnderlineComboBox.ResetItemHeight;
Var ih:Integer;
begin
  ih:=GetItemHeight(Font);
  if ih<9 then ItemHeight:=9 else ItemHeight:=ih;
end;

procedure TUnderlineComboBox.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;



procedure Register;
begin
  RegisterComponents('DeadSouls', [TUnderlineComboBox]);
end;

end.
