unit PlanHash;

interface

Uses Windows,SysUtils;

Type
     PHashRec=^THashRec;
     THashRec=packed record
       nodeid:Integer;
       parentid:Integer;
       childs:Integer;
       nodeNum:ShortString;
       next:PHashRec;
     end;

     TPlanHash=class
      private
       FHashRoot:PHashRec;
       procedure FreeHash;
       function  GetParent(parentid:Integer):PHashRec;
       procedure AppendNode(parentid,nodeid:Integer;nodeNum:ShortString);
       function  GetCountWithThisParent(parentid:Integer):Integer;
      public
       constructor Create;
       destructor  Destroy; override;
       procedure AddNode(parentid,nodeid:Integer);
       function  GetNodeNumber(nodeid:Integer):ShortString;
     end;

implementation

constructor TPlanHash.Create;
 begin
   inherited Create;
   FHashRoot:=nil;
 end;

destructor  TPlanHash.Destroy;
 begin
   if (FHashRoot<>nil) then FreeHash;
   inherited Destroy;
 end;

procedure TPlanHash.FreeHash;
var r1,r2:PHashRec;
 begin
   if FHashRoot=nil then Exit;
   r1:=FHashRoot;
    while (r1<>nil) do
     begin
       r2:=r1;
       r1:=r1^.next;
       FreeMem(r2);
     end;
 end;



function  TPlanHash.GetNodeNumber(nodeid:Integer):ShortString;
var r:PHashRec;
    flag:Boolean;
 begin
   Result:='';
   if FHashRoot=nil then Exit;
   r:=FHashRoot;
   Flag:=True;
   while ((r<>nil) and (flag)) do
    begin
      if (r^.nodeid=nodeid) then
        begin
          Flag:=False;
          Result:=r^.nodeNum;
        end
      else r:=r^.next;
    end;
 end;

function  TPlanHash.GetParent(parentid:Integer):PHashRec;
Var r:PHashRec;
    flag:Boolean;
 begin
   r:=FHashRoot;
   flag:=True;
   while ((r<>nil) and (flag)) do
    begin
      if (r^.nodeid=parentid) then flag:=false
       else
        r:=r^.next;
    end;
   if flag then Result:=nil else Result:=r;
 end;

procedure TPlanHash.AddNode(parentid,nodeid:Integer);
var r:PHashRec;
 begin
  {}
  if FHashRoot=nil then
   begin
     { Add first node with index '1' }
     FHashRoot:=PHashRec(AllocMem(sizeof(THashRec)));
     FHashRoot^.nodeid:=nodeid;
     FHashRoot^.parentid:=parentid;
     FHashRoot^.nodeNum:='1';
     FHashRoot^.childs:=0;
     FHashRoot^.next:=nil;
   end
  else
   begin
    { Get parent node string and }
     r:=GetParent(parentid);
     if (r<>nil) then
      begin
        r^.childs:=r^.childs+1;
        AppendNode(parentid,nodeid,r^.nodeNum+'.'+IntToStr(r^.childs));
      end
     else
      begin
        AppendNode(parentid,nodeid,IntToStr(GetCountWithThisParent(parentid)+1));
      end;
   end;
 end;

procedure TPlanHash.AppendNode(parentid,nodeid:Integer;nodeNum:ShortString);
Var r,r1:PHashRec;
 begin
   {}
   r:=FHashRoot;
   while (r^.next<>nil) do r:=r^.next;
   r1:=PHashRec(AllocMem(sizeof(THashRec)));
   r1^.nodeid:=nodeid;
   r1^.parentid:=parentid;
   r1^.nodeNum:=nodeNum;
   r1^.childs:=0;
   r1^.next:=nil;
   r^.next:=r1;
 end;

function  TPlanHash.GetCountWithThisParent(parentid:Integer):Integer;
var r:PHashRec;
 begin
  {}
  Result:=0;
  r:=FHashRoot;
  while (r<>nil) do
   begin
     if (r^.parentid=parentid) then Result:=Result+1;
     r:=r^.next;
   end;
 end;

end.
