Unit QRRich2;

Interface
uses messages, windows, classes, controls, stdctrls,sysutils, graphics, buttons,
     forms, extctrls, dialogs, printers, db, DBtables, ComCtrls, RichEdit2,
     QRPrntr, Quickrpt, QR2Const;

Type
{ TQRRichEdit - TQRCustomRichEdit descendant with some special settings }
  TCharRange = record
    cpMin: Longint;
    cpMax: LongInt;
  end;


  TFormatRange = record
    hdc: HDC;
    hdcTarget: HDC;
    rc: TRect;
    rcPage: TRect;
    chrg: TCharRange;
  end;


  TQRRichEdit98 = class(TCustomRichEdit98)
  public
    property BorderStyle;
  end;

  { TQRRichText }
  TQRCustomRichText98 = class(TQRPrintable)
  private
    FAutoStretch : boolean;
    FParentRichEdit : TRichEdit98;
    FRichEdit : TQRRichEdit98;
    function GetAlignment : TAlignment;
    function GetColor : TColor;
    function GetFont : TFont;
    function GetLines : TStrings;
    procedure SetAlignment(Value : TAlignment); override;
    procedure SetColor(Value : TColor);
    procedure SetFont(Value : TFont);
    procedure SetLines(Value : TStrings);
    procedure SetParentRichEdit(Value : TRichEdit98);
  protected
    property Lines : TStrings read GetLines write SetLines;
    property ParentRichEdit : TRichEdit98 read FParentRichEdit write SetParentRichEdit;
    procedure Print(OfsX, OfsY : integer); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight : integer); override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    property Alignment : TAlignment read GetAlignment write SetAlignment;
    property AutoStretch : boolean read FAutoStretch write FAutoStretch;
    property Color : TColor read GetColor write SetColor;
    property Font : TFont read GetFont write SetFont;
  end;

  { TQRRichText }
  TQRRichText98 = class(TQRCustomRichText98)
  published
    property Lines;
    property ParentRichEdit;
  end;

  { TQRDBRichText }
  TQRDBRichText98 = class(TQRCustomRichText98)
  private
    Field : TField;
    FDataField : string;
    FDataSet : TDataSet;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetDataSet(Value : TDataSet);
    procedure Prepare; override;
    procedure Unprepare; override;
    procedure Print(OfsX, OfsY : integer); override;
  published
    property DataField : string read FDataField write FDataField;
    property DataSet : TDataSet read FDataSet write SetDataSet;
  end;

Const
  EM_FORMATRANGE                      = WM_USER + 57;
  EM_DISPLAYBAND                      = WM_USER + 51;

procedure Register;

Implementation

{ TQRCustomRichText }

constructor TQRCustomRichText98.Create(AOwner : TComponent);
begin
  FRichEdit := nil;
  inherited Create(AOwner);
  FRichEdit := TQRRichEdit98.Create(Self);
  FRichEdit.Parent := self;
  FRichEdit.BorderStyle := forms.bsNone;
  AutoStretch := false;
  Width := 100;
  Height := 100;
end;

destructor TQRCustomRichText98.Destroy;
begin
  FRichEdit.Free;
  inherited Destroy;
end;

function TQRCustomRichText98.GetAlignment : TAlignment;
begin
  result := FRichEdit.Alignment;
end;

function TQRCustomRichText98.GetColor : TColor;
begin
  result := FRichEdit.Color;
end;

function TQRCustomRichText98.GetFont : TFont;
begin
  result := FRichEdit.Font;
end;

function TQRCustomRichText98.GetLines : TStrings;
begin
  result := FRichEdit.Lines;
end;

procedure TQRCustomRichText98.Print(OfsX, OfsY : integer);
var
  Range: TFormatRange;
  LastChar, MaxLen, LogX, LogY: Integer;
  ARichEdit : TCustomRichEdit98;
  ChangeX, Y : extended;
  VSize : extended;
  OldMapMode : integer;
  i:Integer;
begin
  if assigned(FParentRichEdit) then
    ARichEdit := ParentRichEdit
  else
    ARichEdit := FRichEdit;
  FillChar(Range, SizeOf(TFormatRange), 0);
  with Range do
  begin
    hdc := ParentReport.QRPrinter.Canvas.Handle;
    hdcTarget := hdc;
    LogX := GetDeviceCaps(hdc, LOGPIXELSX);
    LogY := GetDeviceCaps(hdc, LOGPIXELSY);
    rc.Left := QRPrinter.XPos(OfsX + Size.Left) * 1440 div LogX;
    rc.Top := QRPrinter.YPos(OfsY + Size.Top) * 1440 div LogY;
    rc.Right := QRPrinter.XPos(OfsX + Size.Width + Size.Left) * 1440 div LogX;
    rc.Bottom := QRPrinter.YPos(OfsY + Size.Height + Size.Top) * 1440 div LogY;
    rcPage := rc;
    LastChar := 0;
    MaxLen:=0;
     for i:=0 to (ARichEdit.Lines.Count-1) do
       MaxLen:=MaxLen+Length(ARichEdit.Lines[i]);
    chrg.cpMax := -1;
    chrg.cpMin := LastChar;
    Y := OfsY + Size.Top;
    OldMapMode := SetMapMode(hdc, MM_TEXT);
    VSize := OfsY + Size.Top + Size.Height;
    if AutoStretch and (Parent is TQRCustomBand) and (MaxLen > 0) then
    begin
      LastChar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 0, Longint(@Range));
      if (LastChar >= MaxLen) then
      begin
        if TQRCustomBand(Parent).CanExpand(50) then
        begin
          TQRCustomBand(Parent).ExpandBand(50, Y, ChangeX);
          VSize := VSize + 50;
          rc.Bottom := QRPrinter.YPos(VSize) * 1440 div LogY;
        end;
        LastChar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 1, Longint(@Range));
        chrg.cpMin := LastChar;
        rc.Top := rc.Bottom;
      end
      else  // We need to expand, see how much we can print on this page...
        while (LastChar < MaxLen) and (LastChar <> -1) do
        begin
          if TQRCustomBand(Parent).CanExpand(50) then
          begin
            TQRCustomBand(Parent).ExpandBand(50, Y, ChangeX);
            VSize := VSize + 50;
            rc.Bottom := QRPrinter.YPos(VSize) * 1440 div LogY;
            LastChar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 0, Longint(@Range));
            if (LastChar >= MaxLen) or (LastChar = -1) then  // room for everything
            begin
              LastChar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 1, Longint(@Range));
              LastChar := MaxLen;
            end;
          end else  // page break is needed, print whatever fits on this page and go on
          begin
            LastChar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 1, Longint(@Range));
            TQRCustomBand(Parent).ExpandBand(50, Y, ChangeX);
            VSize := Y + 50;
            rc.Top := QRPrinter.YPos(Y) * 1440 div LogY;
            rc.Bottom := QRPrinter.YPos(VSize) * 1440 div LogY;
            rc.Left := rc.Left + QRPrinter.XSize(ChangeX) * 1440 div LogX;
            rc.Right := rc.Right + QRPrinter.XSize(ChangeX) * 1440 div LogX;
            chrg.cpMin := LastChar;
            hdc := ParentReport.QRPrinter.Canvas.Handle;
            hdcTarget := hdc;
           end;
        end;
    end else
    begin
      lastchar := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 1, Longint(@Range));
      inherited;
    end;
  end;
  SetMapMode(ParentReport.QRPrinter.Canvas.Handle, OldMapMode);
  SendMessage(FRichEdit.Handle, EM_FORMATRANGE, 0, 0);
end;

procedure TQRCustomRichText98.SetAlignment(Value : TAlignment);
begin
  FRichEdit.Alignment := Value;
end;

procedure TQRCustomRichText98.SetColor(Value : TColor);
begin
  FRichEdit.Color := Value;
end;

procedure TQRCustomRichText98.SetFont(Value : TFont);
begin
  FRichEdit.Font := Value;
end;

procedure TQRCustomRichText98.SetBounds(ALeft, ATop, AWidth, AHeight : integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if FRichEdit <> nil then FRichEdit.SetBounds(1, 1, AWidth - 2, AHeight - 2);
end;

procedure TQRCustomRichText98.SetLines(Value : TStrings);
begin
  FRichEdit.Lines := Value;
  if assigned(FParentRichEdit) then
    FParentRichEdit.Lines := Value;
end;

procedure TQRCustomRichText98.SetParentRichEdit(Value : TRichEdit98);
begin
  FParentRichEdit := Value;
  if Value<>nil then
    FRichEdit.Lines := Value.Lines;
end;

{ TQRDBRichText }

procedure TQRDBRichText98.SetDataSet(Value : TDataSet);
begin
  FDataSet := Value;
  if Value<>nil then
    Value.FreeNotification(self);
end;

procedure TQRDBRichText98.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if AComponent=FDataSet then
      FDataSet := nil;
end;

procedure TQRDBRichText98.Prepare;
begin
  inherited Prepare;
  if assigned(FDataSet) then
  begin
    Field := FDataSet.FindField(FDataField);
    if (Field is TBlobField) or (Field is TMemoField) then
    begin
      Caption := '';
    end;
  end else
    Field := nil;
end;

procedure TQRDBRichText98.Print(OfsX, OfsY : integer);
begin
  if assigned(Field) then
    if (Field is TMemoField) or
       (Field is TBlobField) then
      Lines.Assign(Field);
  inherited Print(OfsX,OfsY);
end;

procedure TQRDBRichText98.Unprepare;
begin
  Field := nil;
  inherited Unprepare;
end;

procedure Register;
begin
  RegisterComponents('QReport',[TQRRichText98, TQRDBRichText98]);
end;



end.