{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 1995, 1996 AO ROSNO             }
{         Copyright (c) 1997 Master-Bank                }
{                                                       }
{*******************************************************}

unit RXCtlReg;

{$D-,L-,S-}

interface

{ Register custom useful controls }

procedure Register;

implementation

{$IFDEF WIN32}
{$R *.D32}
{$ELSE}
{$R *.D16}
{$ENDIF}

{$I RX.INC}

uses Classes, SysUtils, DsgnIntf, Controls, Graphics, ExtCtrls, Tabs, Dialogs,
  Forms, {$IFDEF RX_D3} DsnConst, {$ELSE} LibConst, {$ENDIF} FiltEdit,
  RXLConst, PictEdit, RXCtrls, CurrEdit, ToolEdit, HintProp, DateUtil,
  PickDate, RXSplit, RXSlider, RXClock, Animate, RxSpin, RXSwitch, RXDice,
  CheckItm, Consts, VCLUtils, RxColors, AniFile, IcoList, IcoLEdit, RxPrgrss;

{$IFNDEF RX_D3}

{ TDateProperty }

type
  TDateProperty = class(TFloatProperty)
  public
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

function TDateProperty.GetValue: string;
begin
  if GetFloatValue = NullDate then Result := ''
  else Result := FormatDateTime(DefDateFormat, GetFloatValue);
end;

procedure TDateProperty.SetValue(const Value: string);
begin
  if Value = '' then SetFloatValue(NullDate)
  else SetFloatValue(StrToDateFmt(DefDateFormat, Value));
end;

{$ENDIF}

{ TPaintBoxEditor }

type
  TPaintBoxEditor = class(TDefaultEditor)
  public
    procedure EditProperty(PropertyEditor: TPropertyEditor;
      var Continue, FreeEditor: Boolean); override;
  end;

procedure TPaintBoxEditor.EditProperty(PropertyEditor: TPropertyEditor;
  var Continue, FreeEditor: Boolean);
begin
  if CompareText(PropertyEditor.GetName, 'OnPaint') = 0 then begin
    PropertyEditor.Edit;
    Continue := False;
  end;
end;

{ TAnimatedEditor }

type
  TAnimatedEditor = class(TComponentEditor)
  private
    procedure EditImage(Image: TAnimatedImage);
    procedure LoadAniFile(Image: TAnimatedImage);
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure TAnimatedEditor.EditImage(Image: TAnimatedImage);
var
  PictureEditor: TPictEditor;
begin
  PictureEditor := TPictEditor.Create(nil);
  try
    PictureEditor.PicDlg.Caption := Image.Name + '.Glyph';
    PictureEditor.GraphicClass := TBitmap;
    PictureEditor.Picture.Graphic := Image.Glyph;
    if PictureEditor.Execute then
      if (PictureEditor.Picture.Graphic = nil) or
         (PictureEditor.Picture.Graphic is PictureEditor.GraphicClass) then
      begin
        Image.Glyph := PictureEditor.Picture.Bitmap;
        Designer.Modified;
      end
      else raise EInvalidGraphicOperation.Create(ResStr(SInvalidPropertyValue));
  finally
    PictureEditor.Free;
  end;
end;

procedure TAnimatedEditor.LoadAniFile(Image: TAnimatedImage);
var
  Dialog: TOpenDialog;
  AniCursor: TAnimatedCursorImage;
begin
  Dialog := TOpenDialog.Create(Application);
  try
    with Dialog do begin
      Options := [ofHideReadOnly, ofFileMustExist];
      DefaultExt := 'ani';
      Filter := LoadStr(srAniCurFilter);
      if Execute then begin
        AniCursor := TAnimatedCursorImage.Create;
        try
          AniCursor.LoadFromFile(FileName);
          AniCursor.AssignToBitmap(Image.Glyph, clFuchsia, True,
            Image.Orientation = goVertical);
          Image.Interval := AniCursor.DefaultRate;
          Image.TransparentColor := clFuchsia;
          Designer.Modified;
        finally
          AniCursor.Free;
        end;
      end;
    end;
  finally
    Dialog.Free;
  end;
end;

procedure TAnimatedEditor.ExecuteVerb(Index: Integer);
begin
  case Index of
    0: EditImage(TAnimatedImage(Component));
    1: LoadAniFile(TAnimatedImage(Component));
  end;
end;

function TAnimatedEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := LoadStr(srEditPicture);
    1: Result := LoadStr(srLoadAniCursor);
  end;
end;

function TAnimatedEditor.GetVerbCount: Integer;
begin
  Result := 2;
end;

{ TProgressControlProperty }

type
  TProgressControlProperty = class(TComponentProperty)
  private
    FProc: TGetStrProc;
    procedure CheckComponent(const AName: string);
  public
    procedure GetValues(Proc: TGetStrProc); override;
  end;

procedure TProgressControlProperty.CheckComponent(const AName: string);
var
  Component: TComponent;
begin
{$IFDEF WIN32}
  Component := Designer.GetComponent(AName);
{$ELSE}
  Component := Designer.Form.FindComponent(AName);
{$ENDIF}
  if (Component <> nil) and (Component is TControl) and
    SupportsProgressControl(TControl(Component)) and Assigned(FProc) then
    FProc(AName);
end;

procedure TProgressControlProperty.GetValues(Proc: TGetStrProc);
begin
  FProc := Proc;
  try
    inherited GetValues(CheckComponent);
  finally
    FProc := nil;
  end;
end;

{ TWeekDayProperty }

type
  TWeekDayProperty = class(TEnumProperty)
    function GetAttributes: TPropertyAttributes; override;
  end;

function TWeekDayProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paMultiSelect, paValueList];
end;

{$IFDEF RX_D3}
resourcestring
  srSamples = 'Samples';
{$ENDIF}

procedure Register;
begin
  RegisterComponents(LoadStr(srRXControls), [TComboEdit, TFilenameEdit,
    TDirectoryEdit, TDateEdit, TCurrencyEdit, TTextListBox, TRxCheckListBox,
    TFontComboBox, TColorComboBox, TRxSplitter, TRxSlider, TRxLabel,
    TRxClock, TAnimatedImage, TRxDrawGrid, TRxSpeedButton, TRxSpinButton,
    TRxSpinEdit, TRxSwitch, TRxDice]);
{$IFDEF CBUILDER}
  RegisterComponents(ResStr(srAdditional), [TScroller]);
{$ELSE}
  RegisterComponents(ResStr(srSamples), [TScroller]);
{$ENDIF}

{$IFDEF RX_D3}
  RegisterNonActiveX([TCustomComboEdit, TCustomDateEdit, TCustomNumEdit,
    TFileDirEdit, TRxCustomListBox], axrComponentOnly);
  RegisterNonActiveX([TScroller], axrComponentOnly);
{$ENDIF RX_D3}

  RegisterPropertyEditor(TypeInfo(TPicture), nil, '', TPictProperty);
  RegisterPropertyEditor(TypeInfo(TGraphic), nil, '', TGraphicPropertyEditor);
  RegisterPropertyEditor(TypeInfo(TDayOfWeekName), nil, '', TWeekDayProperty);
  RegisterPropertyEditor(TypeInfo(string), TCustomNumEdit, 'Text', TStringProperty);
  RegisterPropertyEditor(TypeInfo(string), TFileDirEdit, 'Text', TStringProperty);
  RegisterPropertyEditor(TypeInfo(string), TCustomDateEdit, 'Text', TStringProperty);
  RegisterPropertyEditor(TypeInfo(string), TFileNameEdit, 'Filter', TFilterProperty);
  RegisterPropertyEditor(TypeInfo(string), TFontComboBox, 'FontName', TFontNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TPersistent, 'Hint', THintProperty);
  RegisterPropertyEditor(TypeInfo(string), TCustomComboEdit, 'ButtonHint', THintProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TRxCheckListBox, 'Items', TCheckItemsProperty);
  RegisterPropertyEditor(TypeInfo(TIconList), nil, '', TIconListProperty);
  RegisterPropertyEditor(TypeInfo(TControl), TComponent, 'Gauge', TProgressControlProperty);
  RegisterPropertyEditor(TypeInfo(TControl), TComponent, 'ProgressBar', TProgressControlProperty);
{$IFNDEF RX_D3}
  RegisterPropertyEditor(TypeInfo(TDateTime), TPersistent, '', TDateProperty);
{$ENDIF}

  RegisterComponentEditor(TPaintBox, TPaintBoxEditor);
  RegisterComponentEditor(TImage, TGraphicsEditor);
  RegisterComponentEditor(TAnimatedImage, TAnimatedEditor);
  RxColors.Register;
end;

end.
