Unit BPStream;
{
(c) . .  , .  1997,2006
}
{$INCLUDE Compmode.inc}
{ Corrected 24-12-98 }
{ Corrected 28-09-98 }
{ Rewritten 30-09-98 }
{ Rewritten 22-12-99 }
{ Rewritten 16-11-00 }

Interface

Uses Windows,SysUtils,BPDef,classes,DBTables;

   {  BPD File structure:
          Header
          NodesTypesInfo
          LinksTypesInfo
          Tree
          Data1
          Data2
          Data3
          Data4
          .....
          OpTree
          OpData1
          OpData2
          OpData3
          .......}


Type

 { TBPFileStream }
    TBPReadType = (rt_500,rt_600);

    PBPFileHeader =^TBPFileHeader;
    TBPFileHeader = packed record
      dwSignature:DWORD;      { File signature }
      dwSize:DWORD;           { Self size - offset to tree root }
      ssModel:ShortString;    { Model name }
      ssCompany:ShortString;  { Company name }
      ssAuthor:ShortString;   { Author }
      dfDateTime:TDateTime;   { Date and Time creating model }
      dwRevision:DWORD;       { Document revision }
      dwRootEntries:DWORD;    { Tree entries count }
      dwNodesTypesCount:DWORD;
      dwLinksTypesCount:DWORD;
      dwOpEntries:DWORD;      { OpTree entries count }
      dwOpTreeOffset:DWORD;
      dwInProcessEntries:DWORD; { InProcess entries count }
      dwInProcessOffset:DWORD;
    end;

    PBPRootEntry =^TBPRootEntry;
    TBPRootEntry = packed record
      dwDataOffset:DWORD;
      dwLevel:Integer;          { Opened level e.q. 123 = 1.2.3 }
      dwParent:Integer;
      ssLevel:String[64];
      ssCaption:ShortString;  { Level Caption }
      ssNotes:ShortString;
      dwRevision:DWORD;       { Shit Revision }
      dwNodesCount:DWORD;     { Nodes count   }
      dwLinksCount:DWORD;     { Links count   }
      dwStatus:TDraftStatus;  { Draft status  }
      dwPageNumber:DWORD;     { Page Number   }
      dwLevelType:DWORD;
      dwNodesFontSize:Integer;
      dbOverrideNodesFS:Boolean;
      dwDNodesFontSize:Integer;
      dbOverrideDNodesFS:Boolean;
      dwDProjectionFontSize:Integer;
      dwInProcessFontSize:Integer;
      dbOverrideInProcessFS:Boolean;
      dwLabelsFontSize:Integer;
      dbOverrideLabelsFS:Boolean;
      AvailableMask:word;
      wModelID,wNBPID:word;
      dwLevelID:DWORD;
      ssReserved:packed array[0..221] of byte;
    end;

    PBPOPEntry=^TBPOpEntry;
    TBPOpEntry= packed record
      dwDataOffset:DWORD;
      dwLevel:Integer;
      dwParent:Integer;
      ssCaption:ShortString;
      dwOpDetailCount:DWORD;
      dwNodeIndex:DWORD;
    end;

      TBPStreamOpenMode = ( stRead,stWrite);

      TBPFileStream = class
       private
        FFileName:String;                { File Name }
        FFile:File;                      { Internal variable }
        FBuf:PByte;                      { Internal variable }
        FFileSize:DWORD;
        FOpened:Boolean;
        FMode:TBPStreamOpenMode;
        FRootEntries:DWORD;
        FNodesTypesCount:DWORD;
        FLinksTypesCount:DWORD;
        FReadType:TBPReadType;
        FOpTreeEntries:DWORD;
        FOpTreeOffset:DWORD;
        FInProcessEntries:DWORD;
        FInProcessOffset:DWORD;
        FMask:word;
        FLevelDB:TTable;
        function    ReadRootEntry(entryNumber:DWORD):TBPRootEntry;
        procedure   WriteInProcessArray(Const data:TList;offset:DWORD);
        function    ReadFileHeader:TBPFileHeader;
        procedure   WriteFileHeader(Value: TBPFileHeader);
        procedure   WriteNodesArray(data:TList;offset:DWORD);
        procedure   WriteLinksArray(data:TList;offset:DWORD);
        { New }
        procedure    ReadNodesTypes(ntTree:TList;Count,offset:DWORD);
        procedure    ReadLinksTypes(ltTree:TList;Count,offset:DWORD);
        procedure    WriteNodesTypes(ntTree:TList;offset:DWORD);
        procedure    WriteLinksTypes(ltTree:TList;offset:DWORD);
        procedure    WriteOpDetail(data:TStringList;offset:DWORD);
        { End new }
        function    WriteTree(bpTree:TList;Model,NBP:word):Boolean;
        function    ReadTree(bpTree:TList;RootEntries:DWORD):Boolean;
        function    WriteOpTree(OpTree:TList):Boolean;
        procedure   WriteRootEntry(Const bpRootEntry:TBPRootEntry;entryNumber:DWORD);
        procedure   WriteOpTreeEntry(Const bpOpEntry:TBPOpEntry;entryNumber:DWORD);

       public
        constructor Create;
        destructor  Destroy; override;
        function    OpenStream(oFileName:String;mode:TBPStreamOpenMode):Boolean;
        function    WriteBP(Const bpHeader:TBPHeader; bpTree,
                      ntTree,ltTree,OpTree,IPArray:TList;Model,NBP,mask:word):Boolean;
        function ReadBP(var bpHeader:TBPHeader;ntTree,ltTree,bpTree:TList):Boolean;
        procedure   CloseStream;
       published
        property FileName:String read FFileName;
        property FileSize:DWORD read FFileSize;
        property RootEntries:DWORD read FRootEntries;
        property OpTreeEntries:DWORD read FOpTreeEntries;
      end;

Const     BPSignature500 = $35445042; { 5.00 }
          BPSignature600 = $36445042; { 5.00 }
Var
    IPMode:integer;

Implementation

uses Variants;
constructor TBPFileStream.Create;
 begin
    inherited Create;
    FReadType:=rt_500;
    FFileName:='';
    FFileSize:=0;
    FOpened:=False;
    FBuf:=nil;
    FMode:=stRead;
    FRootEntries:=0;
    FOpTreeEntries:=0;
    FOpTreeOffset:=0;
    FNodesTypesCount:=0;
    FLinksTypesCount:=0;
    FInProcessEntries:=0;
    FInProcessOffset:=0;
    FLevelDB:=TTable.create(nil);
    FLevelDB.tableName:='BPView_New\Levels.db';
 end;

destructor  TBPFileStream.Destroy;
 begin
   if FOpened then CloseStream;
   FLevelDB.Free;
   inherited Destroy;
 end;
procedure   TBPFileStream.WriteInProcessArray(Const data:TList;offset:DWORD);
Var  tmpPtr:PBPInProcessNode;
     nodePtr:PByte;
     i:integer;
begin

   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpptr:=PBPInProcessNode(nodeptr);
   for i:=0 to data.count-1 do
   begin
   System.Move(PBPInProcessNode(data[i])^,tmpPtr^,sizeof(TBPInProcessNode));
   inc(tmpptr);
   end;
 end;

function    TBPFileStream.OpenStream(oFileName:String;mode:TBPStreamOpenMode):Boolean;
 begin
 FFileName:=oFileName;
 Result:=False;
 AssignFile(FFile,FFileName);
 {$I-}
 FMode:=mode;
 case (mode) of
  stRead: Reset(FFile,1);
  stWrite: Rewrite(FFile,1);
 end;
 {$I+}
 if IOResult<>0 then Exit;
 FFileSize:=System.FileSize(FFile);
 if mode = stRead then
     begin
       if (FBuf<>nil) then FreeMem(FBuf);
       FBuf:=PByte(AllocMem(FFileSize));
       System.Seek(FFile,0);
       BlockRead(FFile,FBuf^,FFileSize);
     end
   else
    begin
     FBuf:=nil;
     FFileSize:=0;
    end;
 FOpened:=True;
 Result:=True;
 end;

procedure   TBPFileStream.CloseStream;
 begin
 if FOpened then
   begin
     if (FMode=stWrite) and (FBuf<>nil) then
        begin
         System.Seek(FFile,0);
         BlockWrite(FFile,FBuf^,FFileSize);
         FreeMem(FBuf);
         FBuf:=nil;
        end;
     if (FMode=stRead) and (FBuf<>nil) then
        begin
          FreeMem(FBuf);
          FBuf:=nil;
        end;
     CloseFile(FFile);
     FOpened:=False;
   end;
 end;


procedure   TBPFileStream.WriteFileHeader(Value: TBPFileHeader);
Var tmpPtr:PBPFileHeader;
 begin
   tmpPtr:=PBPFileHeader(FBuf);
   System.Move(Value,tmpPtr^,sizeOf(TBPFileHeader));
 end;


procedure  TBPFileStream.WriteNodesTypes(ntTree:TList;offset:DWORD);
Var tmpPtr:PBPNodeTypes;
    nodePtr:PByte;
    i:integer;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNodeTypes(nodePtr);
   for i:=0 to ntTree.count-1 do
   begin
   System.Move(ntTree[i]^,tmpPtr^,sizeof(TBPNodeTypes));
   inc(tmpPtr,1);
   end;
 end;


procedure  TBPFileStream.WriteLinksTypes(ltTree:TList;offset:DWORD);
Var tmpPtr:PBPLinkTypes;
    nodePtr:PByte;
    i:integer;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLinkTypes(nodePtr);
   for i:=0 to ltTree.count-1 do
   begin
   System.Move(ltTree[i]^,tmpPtr^,sizeof(TBPLinkTypes));
   inc(tmpPtr,1);
   end;
 end;


procedure TBPFileStream.WriteOpDetail(data:TStringList;offset:DWORD);
Var tmpPtr:PBPOpDetail;
    nodePtr:PByte;
    i:integer;
    kind:integer;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPOpDetail(nodePtr);
   for i:=0 to data.count-1 do
   begin
   kind:=Integer(data.Objects[i]);
   {case Kind and 3 of
   0:tmpPtr^.ssCaption:='!'+data[i];
   1:tmpPtr^.ssCaption:=data[i]+'>';
   2:tmpPtr^.ssCaption:='>'+data[i];
   3:tmpPtr^.ssCaption:='>'+data[i]+'>';
   end;}
   tmpPtr^.ssCaption:=data[i];
   inc(tmpPtr,1);
   end;
 end;



procedure   TBPFileStream.WriteNodesArray(data:TList;offset:DWORD);
Var  tmpPtr:PBPNode;
     nodePtr:PByte;
     i:integer;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNode(nodePtr);
   for i:=0 to data.count-1 do
   begin
   System.Move(data[i]^,tmpPtr^,sizeof(TBPNode));
   inc(tmpPtr,1);
   end;
end;
function FLinkOrder(l1,l2:Pointer):integer;
begin
if (PBPLink(l1).dwFromNodeNumber and BPGlobalMask =BPGlobalNodeOperation)and
   (PBPLink(l2).dwToNodeNumber and BPGlobalMask =BPGlobalNodeOperation) then result:=-1
else
if (PBPLink(l1).dwToNodeNumber and BPGlobalMask =BPGlobalNodeOperation)and
   (PBPLink(l2).dwFromNodeNumber and BPGlobalMask =BPGlobalNodeOperation) then result:=+1
else
if (PBPLink(l1).dwFromNodeNumber and BPGlobalMask<>0)and(PBPLink(l1).dwFromNodeNumber and BPGlobalMask<>0) then
 if PBPLink(l1).dwToNodeNumber<PBPLink(l2).dwToNodeNumber then result:=-1
else
if PBPLink(l1).dwToNodeNumber>PBPLink(l2).dwToNodeNumber then result:=+1
else
if PBPLink(l1).dwToLinkNumber<PBPLink(l2).dwToLinkNumber then result:=-1
else
if PBPLink(l1).dwToLinkNumber>PBPLink(l2).dwToLinkNumber then result:=+1
else
result:=0
else
if (PBPLink(l1).dwFromNodeNumber and BPGlobalMask=0)and(PBPLink(l1).dwFromNodeNumber and BPGlobalMask<>0) then
result:=+1
else
if (PBPLink(l1).dwFromNodeNumber and BPGlobalMask<>0)and(PBPLink(l1).dwFromNodeNumber and BPGlobalMask=0) then
result:=-1
else
if PBPLink(l1).dwFromNodeNumber<PBPLink(l2).dwFromNodeNumber then result:=-1
else
if PBPLink(l1).dwFromNodeNumber>PBPLink(l2).dwFromNodeNumber then result:=+1
else
if PBPLink(l1).dwFromLinkNumber<PBPLink(l2).dwFromLinkNumber then result:=-1
else
if PBPLink(l1).dwFromLinkNumber>PBPLink(l2).dwFromLinkNumber then result:=+1
else
result:=0;
end;
procedure   TBPFileStream.WriteLinksArray(data:TList;offset:DWORD);
Var tmpPtr:PSaveBPLink;
    nodePtr:PByte;
    i:integer;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PSaveBPLink(nodePtr);
   data.Sort(flinkorder);
   for i:=0 to data.count-1 do
   begin
   tmpptr^.ssFromCaption:=PBPLink(data[i])^.ssFromCaption;
   tmpptr^.sstoCaption:=PBPLink(data[i])^.sstoCaption;
   tmpptr^.ssGlobalCaption:=PBPLink(data[i])^.ssGlobalCaption;
   tmpptr^.dwFromNodeNumber:=PBPLink(data[i])^.dwFromNodeNumber;
   tmpptr^.dwToNodeNumber:=PBPLink(data[i])^.dwToNodeNumber;
   tmpptr^.IP_SRC:=PBPLink(data[i])^.IP_SRC{ and $FFFF};
   tmpptr^.IP_DST:=PBPLink(data[i])^.IP_DST{ and $FFFF};
   tmpptr^.dwFromLinkNumber:=PBPLink(data[i])^.dwFromLinkNumber;
   tmpptr^.dwToLinkNumber:=PBPLink(data[i])^.dwToLinkNumber;
   tmpptr^.dwFlags:=PBPLink(data[i])^.dwFlags;
   if tmpptr^.ssFromCaption='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_FROMCAPTION
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_SHOW_FROMCAPTION;
   if tmpptr^.ssToCaption='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_TOCAPTION
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_SHOW_TOCAPTION;

   tmpptr^.dwLinkIndex:=PBPLink(data[i])^.dwLinkIndex;
  // System.Move(data[i]^,tmpPtr^,sizeof(TBPLink));
   tmpptr^.ssFromPath:=PBPLink(data[i])^.ssFromPath;
   tmpptr^.ssToPath:=PBPLink(data[i])^.ssToPath;
   if tmpptr^.ssFromPath='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_HAS_FROMPATH
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_HAS_FROMPATH;
   if tmpptr^.ssToPath='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_HAS_TOPATH
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_HAS_TOPATH;

   if ipMode=0 then
   begin
   if (tmpptr^.IP_SRC<>0){and(PBPLink(data[i])^.dwFlags and BP_FROMTOMASK=BP_RIGHT_LEFT)} then
   begin
    tmpptr^.dwFromNodeNumber:=tmpptr^.IP_SRC or BPGlobalNodeInProcess;
    //tmpptr^.dwFromLinkNumber:=1;
{    if (tmpptr^.ssToCaption='')or(tmpptr^.ssToCaption[1] in ['<','[']) then
     tmpptr^.ssToCaption:=tmpPtr.ssFromCaption+tmpptr^.ssToCaption;
    tmpptr^.ssFromCaption:='';
    if tmpptr^.ssToPath='' then
     tmpptr^.ssToPath:=tmpptr^.ssFromPath;
    tmpptr^.ssFromPath:='';
    tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_FROMCAPTION;}
   end;

   if (tmpptr^.IP_DST<>0){and(PBPLink(data[i])^.dwFlags and BP_FROMTOMASK=BP_RIGHT_LEFT)} then
   begin
    tmpptr^.dwToNodeNumber:=tmpptr^.IP_DST or BPGlobalNodeInProcess;
    //tmpptr^.dwToLinkNumber:=1;
{    if (tmpptr^.ssFromCaption='')or(tmpptr^.ssFromCaption[1] in ['<','[']) then
     tmpptr^.ssFromCaption:=tmpPtr.ssToCaption+tmpptr^.ssFromCaption;
    tmpptr^.ssToCaption:='';
    if tmpptr^.ssFromPath='' then
     tmpptr^.ssFromPath:=tmpptr^.ssToPath;
    tmpptr^.ssToPath:='';
    tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_TOCAPTION;}

   end;
   tmpptr^.ssFromCaption:=PBPLink(data[i])^.ssUnMovedFrom+tmpptr^.ssFromCaption;
   tmpptr^.sstoCaption:=tmpptr^.sstoCaption+PBPLink(data[i])^.ssUnMovedTo;
   if tmpptr^.ssFromCaption='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_FROMCAPTION
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_SHOW_FROMCAPTION;
   if tmpptr^.ssToCaption='' then tmpptr^.dwFlags:=tmpptr^.dwFlags and not BP_SHOW_TOCAPTION
                               else tmpptr^.dwFlags:=tmpptr^.dwFlags or BP_SHOW_TOCAPTION;

   end;
   if tmpptr^.IP_SRC<>0 then tmpptr^.dwFlags:=tmpptr^.dwFlags and NOT BP_NOTUSED_TO;
   if tmpptr^.IP_DST<>0 then tmpptr^.dwFlags:=tmpptr^.dwFlags and NOT BP_NOTUSED_FROM;
   inc(tmpPtr,1);
   end;
end;




function    TBPFileStream.WriteBP(Const bpHeader:TBPHeader; bpTree,
                                   ntTree,ltTree,OpTree,IPArray:TList;Model,NBP,mask:word):Boolean;
Var tmp:TBPFileHeader;
    tmpTree:PBPTree;
    totalNodes:DWORD;
    totalLinks:DWORD;
    totalOp:DWORD;
    tmpOffset:DWORD;
    tmpOpTree:PBPOpTree;
    ipoffset:DWORD;
    i:integer;
 begin
   Result:=False;
   fmask:=mask;
   if FBuf<>nil then FreeMem(FBuf);
   FillChar(tmp,sizeof(TBPFileHeader),0);
   tmp.dwRootEntries:=0;
   tmp.dwOpEntries:=0;
   totalNodes:=0;
   totalLinks:=0;

for i:=0 to bpTree.count-1 do
    begin
    tmpTree:=bpTree[i];
     inc(tmp.dwRootEntries);
     totalNodes:=totalNodes+tmpTree^.dwNodesCount;
     totalLinks:=totalLinks+tmpTree^.dwLinksCount;
    end;
   totalOp:=0;
   for i:=0 to OpTree.count-1 do
    begin
       tmpOpTree:=OpTree[i];
      inc(tmp.dwOpEntries);
      totalOp:=totalOp+tmpOpTree^.dwOpDetailCount;
    end;
   FFileSize:=sizeof(TBPFileHeader)
              +tmp.dwRootEntries*sizeof(TBPRootEntry)
              +totalNodes*sizeof(TBPNode)
              +totalLinks*sizeof(TSaveBPLink)+
              bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
              bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes)+
              tmp.dwOpEntries*sizeof(TBPOpEntry)+
              totalOp*sizeof(TBPOpDetail)+bpHeader.dwIPEntries*sizeof(TBPInProcessNode);
IPOFFSET:=sizeof(TBPFileHeader)
              +tmp.dwRootEntries*sizeof(TBPRootEntry)
              +totalNodes*sizeof(TBPNode)
              +totalLinks*sizeof(TSaveBPLink)+
              bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
              bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes)+
              tmp.dwOpEntries*sizeof(TBPOpEntry)+
              totalOp*sizeof(TBPOpDetail);
   FOpTreeOffset:=sizeof(TBPFileHeader)
                 +tmp.dwRootEntries*sizeof(TBPRootEntry)
                 +totalNodes*sizeof(TBPNode)
                 +totalLinks*sizeof(TSaveBPLink)+
                 bpHeader.dwNodesTypesCount*sizeof(TBPNodeTypes)+
                 bpHeader.dwLinksTypesCount*sizeof(TBPLinkTypes);
   FBuf:=PByte(AllocMem(FFileSize));
   if FBuf=nil then Exit;
   tmp.dwSignature:=BPSignature500;
   tmp.dwSize:=sizeof(TBPFileHeader);
   tmp.ssModel:= bpHeader.ssModel;
   tmp.ssCompany:= bpHeader.ssCompany;
   tmp.ssAuthor:= bpHeader.ssAuthor;
   tmp.dfDateTime:= bpHeader.dfDateTime;
   tmp.dwRevision:= bpHeader.dwRevision;
   tmp.dwInProcessEntries:=bpHeader.dwIPEntries;
   tmp.dwInProcessOffset:=ipoffset;
   tmp.dwNodesTypesCount:= bpHeader.dwNodesTypesCount;
   tmp.dwLinksTypesCount:= bpHeader.dwLinksTypesCount;
   tmp.dwOpTreeOffset:= FOpTreeOffset;
   FNodesTypesCount:=bpHeader.dwNodesTypesCount;
   FLinksTypesCount:=bpHeader.dwLinksTypesCount;
   FOpTreeEntries:=tmp.dwOpEntries;
   WriteFileHeader(tmp);
   tmpOffset:=sizeOf(TBPFileHeader);
   WriteNodesTypes(ntTree,tmpOffset);
   tmpOffset:=sizeOf(TBPFileHeader)+FNodesTypesCount*sizeof(TBPNodeTypes);
   WriteLinksTypes(ltTree,tmpOffset);
   fLevelDB.Open;
   try
   Result:=WriteTree(bpTree,Model,NBP);
   finally
   flevelDB.Close;
   end;
   if (FOpTreeEntries>0) then Result:=Result and WriteOpTree(OpTree);
   WriteInProcessArray(IPArray,ipoffset);
 end;

function    TBPFileStream.WriteOpTree(OpTree:TList):Boolean;
Var tmpTree:PBPOpTree;
    tmp:TBPOpEntry;
    i:Integer;
    offset:DWORD;

 begin
  Result:=False;
   i:=0;
   offset:=FOpTreeOffset+sizeof(TBPOpEntry)*opTree.count;
   for i:=0 to OpTree.count-1 do
    begin
    tmpTree:=opTree[i];
     FillChar(tmp,sizeof(TBPOpEntry),0);
     tmp.dwDataOffset:=offset;
     tmp.dwLevel:=tmpTree^.dwLevel;
     tmp.dwParent:=tmpTree^.dwParent;
     case tmpTree^.btFlags and 3 of
     0:tmp.ssCaption:='!'+tmpTree^.ssCaption+'!';
     1:tmp.ssCaption:=tmpTree^.ssCaption+'>';
     2:tmp.ssCaption:='>'+tmpTree^.ssCaption;
     3:tmp.ssCaption:=tmpTree^.ssCaption;
     end;
     //tmp.ssCaption:=tmpTree^.ssCaption;
     tmp.dwOpDetailCount:=tmpTree^.dwOpDetailCount;
     //tmp.btFlags:=tmpTree^.btFlags;
     tmp.dwNodeIndex:=tmpTree^.dwNodeIndex;
     WriteOpTreeEntry(tmp,i);
     WriteOpDetail(tmpTree^.lpOpDetail,offset);
     offset:=offset+tmpTree^.dwOpDetailCount*sizeof(TBPOpDetail);
    end;
  Result:=True;
 end;

function    TBPFileStream.WriteTree( bpTree:TList;Model,NBP:Word):Boolean;
Var tmpTree:PBPTree;
    tmp:TBPRootEntry;
    i:Integer;
    offset:DWORD;
 begin
  Result:=False;
   i:=0;
   offset:=sizeof(TBPFileHeader)+FNodesTypesCount*sizeOf(TBPNodeTypes)+
                                 FLinksTypesCount*sizeOf(TBPLinkTypes)+
                                 sizeof(TBPRootEntry)*bpTree.count;
for i:=0 to bpTree.count-1 do
    begin
    tmpTree:=bpTree[i];
     FillChar(tmp,sizeof(TBPRootEntry),0);
     tmp.dwDataOffset:=offset;
     tmp.dwLevel:=tmpTree^.dwLevel;
     tmp.dwParent:=tmpTree^.dwParent;
     tmp.ssLevel:=tmpTree^.ssLevel;
     tmp.ssCaption:=tmpTree^.ssCaption;
     tmp.ssNotes:=tmpTree^.ssNotes;
     tmp.dwRevision:=tmpTree^.dwRevision;
     tmp.dwNodesCount:=tmpTree^.dwNodesCount;
     tmp.dwLinksCount:=tmpTree^.dwLinksCount;
     tmp.dwStatus:=tmpTree^.dwStatus;
     tmp.dwPageNumber:=i+1;
     tmp.dwLevelType:=0;
     tmp.dwNodesFontSize:=12;
     tmp.dwDNodesFontSize:=10;
     tmp.dwDProjectionFontSize:=8;
     tmp.dwInProcessFontSize:=10;
     tmp.dwLabelsFontSize:=8;
     tmp.dbOverrideNodesFS:=false;
     tmp.dbOverrideDNodesFS:=false;
     tmp.dbOverrideInProcessFS:=false;
     tmp.dbOverrideLabelsFS:=false;
     tmp.wModelID:=Model;
     tmp.wNBPID:=NBP;
     tmp.dwLevelID:=tmpTree^.dwID;
     if flevelDB.Locate('Model;NBP;Level',VarArrayOf([Model,NBP,tmpTree^.dwID]),[]) then
     begin
     tmp.dwNodesFontSize:=FlevelDB.fieldByName('NodesFontSize').AsInteger;
     tmp.dwDNodesFontSize:=FlevelDB.fieldByName('DNodesFontSize').AsInteger;
     tmp.dwDProjectionFontSize:=FlevelDB.fieldByName('DProjectionFontSize').AsInteger;
     tmp.dwInProcessFontSize:=FlevelDB.fieldByName('InProcessFontSize').AsInteger;
     tmp.dwLabelsFontSize:=FlevelDB.fieldByName('LabelsFontSize').AsInteger;
     tmp.dbOverrideNodesFS:=(FLevelDB.FieldByName('Flags').asInteger and 1<>0);
     tmp.dbOverrideDNodesFS:=(FLevelDB.FieldByName('Flags').asInteger and 2<>0);
     tmp.dbOverrideInProcessFS:=(FLevelDB.FieldByName('Flags').asInteger and 4<>0);
     tmp.dbOverrideLabelsFS:=(FLevelDB.FieldByName('Flags').asInteger and 8<>0);

     end;
     tmp.AvailableMask:=fmask;
     WriteRootEntry(tmp,i);
     WriteNodesArray(tmpTree^.lpNodes,offset);
     offset:=offset+tmpTree^.dwNodesCount*sizeof(TBPNode);
     WriteLinksArray(tmpTree^.lpLinks,offset);
     offset:=offset+tmpTree^.dwLinksCount*sizeof(TSaveBPLink);
    end;
  Result:=True;
 end;

procedure   TBPFileStream.WriteRootEntry(Const bpRootEntry:TBPRootEntry;entryNumber:DWORD);
Var  tmpPtr:PBPRootEntry;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,sizeof(TBPFileHeader)+FNodesTypesCount*sizeOf(TBPNodeTypes)+
                                 FLinksTypesCount*sizeOf(TBPLinkTypes)+entryNumber*sizeof(TBPRootEntry));
   tmpPtr:=PBPRootEntry(nodePtr);
   System.Move(bpRootEntry,tmpPtr^,sizeof(TBPRootEntry));
 end;

procedure   TBPFileStream.WriteOpTreeEntry(Const bpOpEntry:TBPOpEntry;entryNumber:DWORD);
Var  tmpPtr:PBPOpEntry;
     nodePtr:PByte;
begin
   nodePtr:=FBuf;
   Inc(nodePtr,FOpTreeOffset+entryNumber*sizeof(TBPOpEntry));
   tmpPtr:=PBPOpEntry(nodePtr);
   System.Move(bpOpEntry,tmpPtr^,sizeof(TBPOpEntry));
 end;
function TBPFileStream.ReadBP(var bpHeader:TBPHeader;ntTree,ltTree,bpTree:TList):Boolean;
Var tmp:TBPFileHeader;
    tmpOffset:DWORD;
 begin
    Result:=False;
    if not(FOpened) then Exit;
    tmp:=ReadFileHeader;
    if (tmp.dwSignature<>BPSignature500)and(tmp.dwSignature<>BPSignature600)    then Exit;
    if (tmp.dwSignature=BPSignature500) then FReadType:=rt_500
                                        else FReadType:=rt_600;
    bpHeader.ssModel   := tmp.ssModel;
    bpHeader.ssCompany := tmp.ssCompany;
    bpHeader.ssAuthor  := tmp.ssAuthor;
    bpHeader.dfDateTime:= tmp.dfDateTime;
    bpHeader.dwRevision:= tmp.dwRevision;
    bpHeader.dwNodesTypesCount:=tmp.dwNodesTypesCount;
    bpHeader.dwLinksTypesCount:=tmp.dwLinksTypesCount;
    bpHeader.dwOpEntries:=tmp.dwOpEntries;
    FNodesTypesCount:=tmp.dwNodesTypesCount;
    FLinksTypesCount:=tmp.dwLinksTypesCount;
    FOpTreeEntries:=tmp.dwOpEntries;
    {ltTree:=TList.create;
    ntTree:=TList.create;}
    tmpOffset:=sizeOf(TBPFileHeader);
    ReadNodesTypes(ntTree,FNodesTypesCount,tmpOffset);
    tmpOffset:=sizeOf(TBPFileHeader)+FNodesTypesCount*sizeof(TBPNodeTypes);
    ReadLinksTypes(ltTree,FLinksTypesCount,tmpOffset);
    FLevelDB.Open;
    try
    Result:=ReadTree(bpTree,tmp.dwRootEntries);
    finally
    FLevelDB.Close;
    end;
 end;
function   TBPFileStream.ReadFileHeader:TBPFileHeader;
Var tmpPtr:PBPFileHeader;
 begin
   tmpPtr:=PBPFileHeader(FBuf);
   System.Move(tmpPtr^,Result,sizeOf(TBPFileHeader));
   FRootEntries:=Result.dwRootEntries;
   FNodesTypesCount:=Result.dwNodesTypesCount;
   FLinksTypesCount:=Result.dwLinksTypesCount;
   FOpTreeEntries:=Result.dwOpEntries;
   FOpTreeOffset:=Result.dwOpTreeOffset;
 end;
procedure  TBPFileStream.ReadNodesTypes(ntTree:TList;Count,offset:DWORD);
Var tmpPtr,temp:PBPNodeTypes;
    nodePtr:PByte;
    var i:integer;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPNodeTypes(nodePtr);
   for i:=0 to count-1 do
   begin
   new(temp);
   System.Move(tmpPtr^,temp^,sizeof(TBPNodeTypes));
   ntTree.add(temp);
   inc(tmpPtr,1);
   end;

 end;
procedure  TBPFileStream.ReadLinksTypes(ltTree:TList;Count,offset:DWORD);
Var tmpPtr,temp:PBPLinkTypes;
    nodePtr:PByte;
    i:integer;
 begin
   nodePtr:=FBuf;
   Inc(nodePtr,offset);
   tmpPtr:=PBPLinkTypes(nodePtr);
   for i:=0 to Count-1 do
   begin
   new(temp);
   System.Move(tmpPtr^,temp^,sizeof(TBPLinkTypes));
   ltTree.add(temp);
   inc(tmpPtr,1);
   end;
 end;
function    TBPFileStream.ReadTree(bpTree:TList;RootEntries:DWORD):Boolean;
Var i:Integer;
    tmp:TBPRootEntry;
    tmpTree,oldTmpTree:PBPTree;
    ModelID,NBPID,LevelID:integer;
 begin
   Result:=False;
   oldTmpTree:=nil;
   for i:=0 to (RootEntries-1) do
    begin
      tmp:=ReadRootEntry(i); { ⠫ । entry }
      tmpTree:=PBPTree(AllocMem(sizeof(TBPTree)));
      tmpTree^.dwDataOffset:=tmp.dwDataOffset;
      tmpTree^.dwLevel:=tmp.dwLevel;
      tmpTree^.dwParent:=tmp.dwParent;
      tmpTree^.ssLevel:=tmp.ssLevel;
      tmpTree^.ssCaption:=tmp.ssCaption;
      tmpTree^.ssNotes:=tmp.ssNotes;
      tmpTree^.dwRevision:=tmp.dwRevision;
      tmpTree^.dwNodesCount:=tmp.dwNodesCount;
      tmpTree^.dwLinksCount:=tmp.dwLinksCount;
      tmpTree^.dwStatus:=tmp.dwStatus;
      tmpTree^.dwPageNumber:=tmp.dwPageNumber;
      tmpTree^.dwNodesMemRequired:=tmp.dwNodesCount*sizeof(TBPNode);
      tmpTree^.dwLinksMemRequired:=tmp.dwLinksCount*sizeof(TBPLink);
      tmpTree^.lpNodes:=TList.Create;
      tmpTree^.lpLinks:=TList.Create;
      ModelID:=tmp.wModelID;
      NBPID:=tmp.wNBPID;
      LevelID:=tmp.dwLevelID;
      if fleveldb<>nil then
      begin
      if not fleveldb.Locate('Model;NBP;Level',VarArrayOf([ModelID,NBPID,LevelID]),[]) then
      begin
      fleveldb.Append;
      flevelDB.FieldByName('Model').asInteger:=ModelID;
      flevelDB.FieldByName('NBP').asInteger:=NBPID;
      flevelDB.FieldByName('Level').asInteger:=LevelID;
      end
      else
      flevelDB.Edit;
      flevelDB.FieldByName('NodesFontSize').asInteger:=tmp.dwNodesFontSize;
      flevelDB.FieldByName('DNodesFontSize').asInteger:=tmp.dwDNodesFontSize;
      flevelDB.FieldByName('DProjectionFontSize').asInteger:=tmp.dwDProjectionFontSize;
      flevelDB.FieldByName('InProcessFontSize').asInteger:=tmp.dwInProcessFontSize;
      flevelDB.FieldByName('LabelsFontSize').asInteger:=tmp.dwLabelsFontSize;
      flevelDB.FieldByName('Flags').asInteger:=
                 (integer(tmp.dbOverrideNodesFS) and 1)or
                ((integer(tmp.dbOverrideDNodesFS) and 1)shl 1) or
                ((integer(tmp.dbOverrideInProcessFS) and 1)shl 2) or
                ((integer(tmp.dbOverrideLabelsFS) and 1)shl 3);

      flevelDB.Post;
      end;
      bpTree.Add(tmptree);
{      ReadNodesArray(tmpTree^.lpNodes,tmpTree^.dwNodesCount,tmpTree^.dwDataOffset);
      ReadLinksArray(tmpTree^.lpLinks,tmpTree^.dwLinksCount,tmpTree^.dwDataOffset+tmpTree^.dwNodesMemRequired);}
    end;
   Result:=True;
 end;
function    TBPFileStream.ReadRootEntry(entryNumber:DWORD):TBPRootEntry;
Var tmpPtr:PBPRootEntry;
    nodePtr:PByte;
    tmpOffset:DWORD;
begin
   nodePtr:=FBuf;
   tmpOffset:=sizeof(TBPFileHeader)+FLinksTypesCount*SizeOf(TBPLinkTypes)+
                                    FNodesTypesCount*SizeOf(TBPNodeTypes);
   Inc(nodePtr,tmpOffset+entryNumber*sizeof(TBPRootEntry));
   tmpPtr:=PBPRootEntry(nodePtr);
   System.Move(tmpPtr^,Result,sizeof(TBPRootEntry));
 end;

end.

